# SPDX-FileCopyrightText: 2025 Florian Obersteiner / KIT
# SPDX-FileContributor: Florian Obersteiner <f.obersteiner@kit.edu>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Misc tools."""

import os
from collections.abc import Callable
from pathlib import Path
from typing import Any

import numpy as np

###############################################################################


class PCOLOR:
    """
    Print in color, using ANSI escape codes.

    This class provides ANSI escape codes for various text colors and styles.
    Usage: print(f"{PCOLOR.red}This text is red{PCOLOR.reset}")

    Caution: not all terminals suppot this.
    """

    # Text colors
    black = "\x1b[30m"
    red = "\x1b[31m"
    green = "\x1b[32m"
    yellow = "\x1b[33m"
    blue = "\x1b[34m"
    magenta = "\x1b[35m"
    cyan = "\x1b[36m"
    white = "\x1b[37m"

    # Bright text colors
    bright_black = "\x1b[90m"
    bright_red = "\x1b[91m"
    bright_green = "\x1b[92m"
    bright_yellow = "\x1b[93m"
    bright_blue = "\x1b[94m"
    bright_magenta = "\x1b[95m"
    bright_cyan = "\x1b[96m"
    bright_white = "\x1b[97m"

    # Background colors
    bg_black = "\x1b[40m"
    bg_red = "\x1b[41m"
    bg_green = "\x1b[42m"
    bg_yellow = "\x1b[43m"
    bg_blue = "\x1b[44m"
    bg_magenta = "\x1b[45m"
    bg_cyan = "\x1b[46m"
    bg_white = "\x1b[47m"

    # Bright background colors
    bg_bright_black = "\x1b[100m"
    bg_bright_red = "\x1b[101m"
    bg_bright_green = "\x1b[102m"
    bg_bright_yellow = "\x1b[103m"
    bg_bright_blue = "\x1b[104m"
    bg_bright_magenta = "\x1b[105m"
    bg_bright_cyan = "\x1b[106m"
    bg_bright_white = "\x1b[107m"

    # Text styles
    bold = "\x1b[1m"
    dim = "\x1b[2m"
    italic = "\x1b[3m"
    underline = "\x1b[4m"
    blink = "\x1b[5m"
    reverse = "\x1b[7m"
    hidden = "\x1b[8m"
    strikethrough = "\x1b[9m"

    # Reset codes
    reset = "\x1b[0m"
    disable = "\x1b[0m"  # Alias for reset...

    @staticmethod
    def rgb(r: int, g: int, b: int):
        """Create a custom RGB color code (foreground)."""
        return f"\x1b[38;2;{r};{g};{b}m"

    @staticmethod
    def bg_rgb(r: int, g: int, b: int):
        """Create a custom RGB color code (background)."""
        return f"\x1b[48;2;{r};{g};{b}m"


###############################################################################


def clean_path(p: Path | str, resolve: bool = True) -> Path:
    """Clean a path from stuff like ~ or $HOME."""
    path_str = Path(p).expanduser().as_posix()  # expanduser() resolves '~' to home path
    if path_str.startswith("$HOME"):
        path_str = path_str.replace("$HOME", Path().home().as_posix(), 1)

    if not resolve:
        return Path(path_str)
    return Path(path_str).resolve()


def to_list_of_Path(folders: str | list[str] | Path | list[Path]) -> list[Path]:
    """Turn input string or list of strings into a list of pathlib.Path objects."""
    if not isinstance(folders, list):
        folders = [folders]  # type: ignore

    return [Path(f) for f in folders]  # type: ignore


def insensitive_pattern(pattern: str) -> str:
    """Return a case-insensitive pattern to use in glob.glob or path.glob."""

    def either(c):
        return f"[{c.lower()}{c.upper()}]" if c.isalpha() else c

    return "".join(map(either, pattern))


###############################################################################


def print_progressbar(
    iteration: int,
    total: int,
    prefix: str = "",
    suffix: str = "",
    decimals: int = 1,
    length: int = 100,
    fill: str = "█",
):
    """
    Print a progress bar.

    https://stackoverflow.com/questions/3173320/text-progress-bar-in-the-console
    access: 2018-12-20
    Call in a loop to create terminal progress bar
    @params:
        iteration   - Required  : current iteration (Int)
        total       - Required  : total iterations (Int)
        prefix      - Optional  : prefix string (Str)
        suffix      - Optional  : suffix string (Str)
        decimals    - Optional  : positive number of decimals in percent complete (Int)
        length      - Optional  : character length of bar (Int)
        fill        - Optional  : bar fill character (Str)
    """
    percent = ("{0:." + str(decimals) + "f}").format(100 * (iteration / float(total)))
    filledLength = int(length * iteration // total)
    progBar = fill * filledLength + "-" * (length - filledLength)
    print(f"\r{prefix} |{progBar}| {percent}%% {suffix}", end="\r")
    # Print New Line on Complete
    if iteration == total:
        print("\n")


# #
# # Sample Usage
# #

# from time import sleep

# # A List of Items
# items = list(range(0, 57))
# l = len(items)

# # Initial call to print 0% progress
# printProgressBar(0, l, prefix = 'Progress:', suffix = 'Complete', length = 50)
# for i, item in enumerate(items):
#     # Do stuff...
#     sleep(0.1)
#     # Update Progress Bar
#     printProgressBar(i + 1, l, prefix = 'Progress:', suffix = 'Complete', length = 50)


###############################################################################


def find_youngest_file(path: Path, pattern: str, n: int = 1) -> list[Path] | None:
    """
    Find the file that matches a pattern and has the youngest modification timestamp if there are multiple files that match the pattern.

    input:
        path, string or pathlib.Path, where to look for the file(s)
        pattern, string, pattern to look for in files (see pathlib.Path.glob)
        n, integer, how many to return. defaults to 1.

    Returns
    -------
        filename(s) of youngest file(s), including path
        None if no file
    """
    assert n >= 1, "n must be greater equal 1."

    path = Path(path)
    files = [Path(f) for f in path.glob(pattern) if os.path.isfile(f)]
    sortfiles = sorted(files, key=lambda x: x.stat().st_mtime, reverse=True)

    if sortfiles:
        return sortfiles[:n]
    return None


###############################################################################


def find_fist_elem(arr: list | np.ndarray, val: Any, condition: Callable) -> Any | None:
    """
    Find the first element in arr that gives (arr[ix] condition val) == True.

    Inputs:
        arr: numeric numpy 1d array or python list
        val: scalar value
        condition: e.g. 'operator.ge' (operator package)

    Returns
    -------
        index of value matching the condition or None if no match is found.
    """
    if isinstance(arr, list):
        return next((ix for ix, v in enumerate(arr) if condition(v, val)), None)

    result = np.argmax(condition(arr, val))

    return result if condition(arr[result], val) else None


###############################################################################


def list_change_elem_index(lst: list, element: Any, new_index: int) -> list:
    """
    Change the index of an element in a list.

    see <https://stackoverflow.com/a/3173159/10197418>
    """
    if new_index >= len(lst) or new_index < 0:
        raise IndexError("new index is out of range")

    _lst = lst.copy()

    if element in lst:
        _lst.insert(new_index, _lst.pop(lst.index(element)))

    return _lst


###############################################################################


def set_compare(a: set, b: set) -> tuple[set, set, set]:
    """
    Compare two iterables a and b. set() is used for comparison, so only unique elements will be considered.

    Parameters
    ----------
    a : set
    b : set

    Returns
    -------
    tuple with 3 elements:
        (what is only in a but not in b,
         what is only in b but not in a,
         what is common in a and b)
    """
    return (a - b, b - a, a.intersection(b))



# SPDX-FileCopyrightText: 2025 Florian Obersteiner / KIT
# SPDX-FileContributor: Florian Obersteiner <f.obersteiner@kit.edu>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import os
import sys
import tempfile
import tomllib
import unittest
from io import StringIO

from pyfuppes.dictstuff import DotDict, compare_dictionaries


class TestDotDict(unittest.TestCase):
    """Test cases for the enhanced DotDict class with TOML support."""

    def setUp(self):
        """Initialize test data and capture stdout for warning messages."""
        self.sample_dict = {
            "server": {"host": "localhost", "port": 8080},
            "database": {
                "credentials": {"username": "admin", "password": "secret123"},
                "settings": {"timeout": 30},
            },
            "users": [{"name": "Alice", "role": "admin"}, {"name": "Bob", "role": "user"}],
            "valid_key": "value",
            "invalid-key": "problematic",
            "123numeric": "starts with number",
            "get": "overrides method",
        }

        # Valid TOML content
        self.valid_toml = """
[server]
host = "localhost"
port = 8080

[database.credentials]
username = "admin"
password = "secret123"

[database.settings]
timeout = 30

[[users]]
name = "Alice"
role = "admin"

[[users]]
name = "Bob"
role = "user"

valid_key = "value"
invalid-key = "problematic"
123numeric = "starts with number"
get = "overrides method"
"""

        # Invalid TOML content (syntax error)
        self.invalid_toml = """
[server]
host = "localhost
port = 8080
"""

        # Create temporary files for TOML testing
        self.valid_toml_file = tempfile.NamedTemporaryFile(  # noqa: SIM115
            delete=False, mode="w", suffix=".toml"
        )
        self.valid_toml_file.write(self.valid_toml)
        self.valid_toml_file.close()

        self.invalid_toml_file = tempfile.NamedTemporaryFile(  # noqa: SIM115
            delete=False, mode="w", suffix=".toml"
        )
        self.invalid_toml_file.write(self.invalid_toml)
        self.invalid_toml_file.close()

        # Setup stdout capture
        self.captured_output = StringIO()
        self.original_stdout = sys.stdout
        sys.stdout = self.captured_output

    def tearDown(self):
        """Clean up temporary files and restore stdout."""
        os.unlink(self.valid_toml_file.name)
        os.unlink(self.invalid_toml_file.name)
        sys.stdout = self.original_stdout

    def test_init_with_dict(self):
        """Test initialization with a dictionary."""
        dot_dict = DotDict(self.sample_dict)

        # Test basic access
        self.assertEqual(dot_dict.server.host, "localhost")
        self.assertEqual(dot_dict.server.port, 8080)
        self.assertEqual(dot_dict.database.credentials.username, "admin")

        # Test list of dicts
        self.assertEqual(dot_dict.users[0].name, "Alice")
        self.assertEqual(dot_dict.users[1].role, "user")

        # Verify warnings were printed
        output = self.captured_output.getvalue()
        self.assertIn("Warning: 'invalid-key'", output)
        self.assertIn("Warning: '123numeric'", output)
        self.assertIn("Warning: 'get'", output)

    def test_load_valid_toml_file(self):
        """Test loading a valid TOML file."""
        # Temporarily reset stdout to avoid capturing import warnings
        sys.stdout = self.original_stdout

        # Restore stdout capture
        sys.stdout = self.captured_output

        # Load TOML file
        dot_dict = DotDict.from_toml_file(self.valid_toml_file.name)  # type: ignore

        # Test basic access
        self.assertEqual(dot_dict.server.host, "localhost")
        self.assertEqual(dot_dict.server.port, 8080)
        self.assertEqual(dot_dict.database.credentials.username, "admin")

        # Test list of dicts
        self.assertEqual(dot_dict.users[0].name, "Alice")
        self.assertEqual(dot_dict.users[1].role, "user")

        # Verify warnings were printed
        output = self.captured_output.getvalue()
        self.assertIn("Warning: 'invalid-key'", output)
        self.assertIn("Warning: '123numeric'", output)
        self.assertIn("Warning: 'get'", output)

    def test_load_invalid_toml_file(self):
        """Test loading an invalid TOML file raises appropriate exception."""
        # Reset and restore stdout to avoid capturing import warnings
        sys.stdout = self.original_stdout
        sys.stdout = self.captured_output

        # Try to load invalid TOML file
        with self.assertRaises(tomllib.TOMLDecodeError):
            DotDict.from_toml_file(self.invalid_toml_file.name)  # type: ignore

    def test_access_invalid_keys(self):
        """Test accessing keys that are not valid Python identifiers."""
        dot_dict = DotDict(self.sample_dict)

        # These should work with dictionary access
        self.assertEqual(dot_dict["invalid-key"], "problematic")
        self.assertEqual(dot_dict["123numeric"], "starts with number")
        self.assertEqual(dot_dict["get"], "overrides method")

        # But fail with attribute access
        with self.assertRaises(AttributeError):
            _ = dot_dict.invalid - key  # noqa # type: ignore

        # This might work in Python, but it's a subtraction operation
        # dot_dict.invalid-key is interpreted as (dot_dict.invalid) - key
        with self.assertRaises(AttributeError):
            _ = dot_dict.invalid

        # with self.assertRaises(NameError):
        #     _ = dot_dict.123numeric  # SyntaxError at compile time

    def test_deep_nesting(self):
        """Test deeply nested structures."""
        deep_dict = {"level1": {"level2": {"level3": {"level4": {"value": "deep"}}}}}

        dot_dict = DotDict(deep_dict)
        self.assertEqual(dot_dict.level1.level2.level3.level4.value, "deep")

    def test_update_nested_attribute(self):
        """Test updating a nested attribute."""
        dot_dict = DotDict(self.sample_dict)

        # Update nested attribute
        dot_dict.server.host = "new-host"
        self.assertEqual(dot_dict.server.host, "new-host")
        self.assertEqual(dot_dict["server"]["host"], "new-host")

        # Update with a new dictionary should convert to DotDict
        dot_dict.server = {"host": "another-host", "new_key": "value"}
        self.assertEqual(dot_dict.server.host, "another-host")  # type: ignore
        self.assertEqual(dot_dict.server.new_key, "value")  # type: ignore

        # Verify it's a DotDict
        self.assertIsInstance(dot_dict.server, DotDict)

    def test_from_toml(self):
        """Test creating DotDict from already parsed TOML data."""
        try:
            # Reset and restore stdout to avoid capturing import warnings
            sys.stdout = self.original_stdout
            with open(self.valid_toml_file.name, "rb") as f:
                toml_data = tomllib.load(f)

            sys.stdout = self.captured_output

            # Create DotDict from parsed TOML
            dot_dict = DotDict.from_toml_data(toml_data)

            # Test access
            self.assertEqual(dot_dict.server.host, "localhost")
            self.assertEqual(dot_dict.users[0].name, "Alice")

        except ImportError:
            self.skipTest("TOML library not available")

    def test_nested_list_handling(self):
        """Test handling of nested lists with dictionaries."""
        complex_dict = {
            "nested_lists": [
                [{"name": "item1"}, {"name": "item2"}],
                [{"name": "item3"}, {"name": "item4"}],
            ]
        }

        dot_dict = DotDict(complex_dict)

        # Test that dictionaries in nested lists are converted
        self.assertEqual(dot_dict.nested_lists[0][0].name, "item1")
        self.assertEqual(dot_dict.nested_lists[1][1].name, "item4")

        # Verify outer list items are DotDict instances
        self.assertIsInstance(dot_dict.nested_lists[0][0], DotDict)


class TestCompareDictionaries(unittest.TestCase):
    def test_empty_dictionaries(self):
        missing, extra, failed = compare_dictionaries({}, {})
        self.assertEqual(missing, [])
        self.assertEqual(extra, [])
        self.assertFalse(failed)

    def test_identical_dictionaries(self):
        ref = {"a": 1, "b": 2}
        cand = {"a": 1, "b": 2}
        missing, extra, failed = compare_dictionaries(ref, cand)
        self.assertEqual(missing, [])
        self.assertEqual(extra, [])
        self.assertFalse(failed)

    def test_missing_key(self):
        ref = {"a": 1, "b": 2}
        cand = {"a": 1}
        missing, extra, failed = compare_dictionaries(ref, cand)
        self.assertEqual(missing, ["b"])
        self.assertEqual(extra, [])
        self.assertTrue(failed)

    def test_extra_key(self):
        ref = {"a": 1}
        cand = {"a": 1, "b": 2}
        missing, extra, failed = compare_dictionaries(ref, cand)
        self.assertEqual(missing, [])
        self.assertEqual(extra, ["b"])
        self.assertFalse(failed)  # No missing keys, so comparison doesn't fail

    def test_nested_dictionaries_missing_key(self):
        ref = {"a": 1, "b": {"c": 2, "d": 3}}
        cand = {"a": 1, "b": {"c": 2}}
        missing, extra, failed = compare_dictionaries(ref, cand)
        self.assertEqual(missing, ["b.d"])
        self.assertEqual(extra, [])
        self.assertTrue(failed)

    def test_nested_dictionaries_extra_key(self):
        ref = {"a": 1, "b": {"c": 2}}
        cand = {"a": 1, "b": {"c": 2, "d": 3}}
        missing, extra, failed = compare_dictionaries(ref, cand)
        self.assertEqual(missing, [])
        self.assertEqual(extra, ["b.d"])
        self.assertFalse(failed)

    def test_nested_dictionaries_different_keys(self):
        ref = {"a": 1, "b": {"c": 2, "d": 3}}
        cand = {"a": 1, "b": {"e": 4, "f": 5}}
        missing, extra, failed = compare_dictionaries(ref, cand)
        self.assertEqual(missing, ["b.c", "b.d"])
        self.assertEqual(extra, ["b.e", "b.f"])
        self.assertTrue(failed)

    def test_complex_nested(self):
        ref = {"a": 1, "b": {"c": 2, "d": {"e": 3, "f": 4}}, "g": 5}
        cand = {"a": 1, "b": {"c": 2, "d": {"e": 3}}, "h": 6}
        missing, extra, failed = compare_dictionaries(ref, cand)
        self.assertEqual(missing, ["b.d.f", "g"])
        self.assertEqual(extra, ["h"])
        self.assertTrue(failed)

    def test_nested_extra_key_within_existing_key(self):
        ref = {"a": 1, "b": {"c": 2}}
        cand = {"a": 1, "b": {"c": 2, "d": 3}}
        missing, extra, failed = compare_dictionaries(ref, cand)
        self.assertEqual(missing, [])
        self.assertEqual(extra, ["b.d"])
        self.assertFalse(failed)

    def test_nested_extra_key_within_existing_key_2(self):
        ref = {"a": 1, "b": {"c": 2, "d": {"e": 3}}}
        cand = {"a": 1, "b": {"c": 2, "d": {"e": 3, "f": 4}, "g": 5}}
        missing, extra, failed = compare_dictionaries(ref, cand)
        self.assertEqual(missing, [])
        self.assertEqual(extra, ["b.d.f", "b.g"])
        self.assertFalse(failed)


if __name__ == "__main__":
    unittest.main()



# SPDX-FileCopyrightText: 2025 Florian Obersteiner / KIT
# SPDX-FileContributor: Florian Obersteiner <f.obersteiner@kit.edu>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Helpers to format plots."""

import numpy as np

###############################################################################


def get_plot_range(
    v: list | np.ndarray | np.ma.masked_array,
    add_percent: float = 5,
    v_min_lim: float | None = None,
    v_max_lim: float | None = None,
    xrange: list[float] | None = None,
    x: list | np.ndarray | None = None,
) -> tuple[float, float]:
    """
    Adjust y-axis range of matplotlib pyplot for a given vector v.

    Parameters
    ----------
    v : list or numpy 1d array
        dependent variable.
    add_percent : numeric type scalar value, optional
        percent of the range of v that should be added to result. The default is 5.
    v_min_lim : numeric type scalar value, optional
        minimum value for lower yrange limit. The default is None.
    v_max_lim : numeric type scalar value, optional
        maximum value for upper yrange limit. The default is None.
    xrange : list, optional
        [lower_limit, upper_limit] of independent variable. The default is None.
    x : list or numpy 1d array, optional
        independent variable. The default is None.

    Returns
    -------
    result : tuple, 2 elements
        lower and upper limit.

    """
    if isinstance(v, np.ma.masked_array):
        v = v[~v.mask]

    if not isinstance(v, np.ndarray):
        v = np.array(v)  # ensure array type

    if x is not None and xrange is not None:
        if not isinstance(x, np.ndarray):
            x = np.array(x)
        x = np.sort(x)  # monotonically increasing x-vector (e.g. time)
        if len(x) == len(v):
            w_xvd = (x >= xrange[0]) & (x <= xrange[1])  # type: ignore
            v = v[w_xvd]

    v = v[np.isfinite(v)]

    if len(v) < 2:
        # it is better not to raise an exception in case no valid input,
        # to avoid errors further down...
        return (-1.0, 1.0)

    v_min, v_max = np.min(v), np.max(v)
    offset = (abs(v_min) + abs(v_max)) / 2 * add_percent / 100
    result = [v_min - offset, v_max + offset]

    if v_min_lim and result[0] < v_min_lim:
        result[0] = v_min_lim

    if v_max_lim and result[1] > v_max_lim:
        result[1] = v_max_lim

    return tuple(result)


###############################################################################


def nticks_yrange(
    yrange: list[float], nticks: int, range_as_multiple_of: int = 10
) -> tuple[float, float]:
    """
    Update a plot yrange so that it fits nicely with a certain number of ticks.

    Parameters
    ----------
    yrange : 2-element tuple or list
        the yrange to modify.
    nticks : int
        number of ticks along y-axis.
    range_as_multiples_of: int, optional
        make the yrange divisible w/o remainder by .... The default is 10.

    Returns
    -------
    result : 2-element tuple
        updated yrange.

    Examples
    --------
    >>> nticks_yrange([12, 47], nticks=5, range_as_multiple_of=10)
    (10.0, 50.0)
    >>> # Explanation:
    >>> # - Input range [12, 47] is first rounded to [10, 50] to be divisible by 10
    >>> # - With nticks=5, we need 4 equal intervals (nticks-1)
    >>> # - The range 40 is divisible by both 4 and 10, so the result is (10.0, 50.0)

    >>> nticks_yrange([12, 47], nticks=6, range_as_multiple_of=5)
    (10.0, 50.0)
    >>> # Explanation:
    >>> # - Input range [12, 47] is rounded to [10, 50] to be divisible by 5
    >>> # - With nticks=6, we need 5 equal intervals (nticks-1)
    >>> # - The range 40 is divisible 5, so the result is (10.0, 50.0)

    >>> nticks_yrange([-3, 22], nticks=7, range_as_multiple_of=10)
    (-10.0, 50.0)
    >>> # Explanation:
    >>> # - Input range [-3, 22] is rounded to [-10, 30] to be divisible by 10
    >>> # - With nticks=7, we need 6 equal intervals (nticks-1)
    >>> # - The LCM of 6 and 10 is 30, so the range needs to be a multiple of 30
    >>> # - The range is expanded to 60 (2 × 30), resulting in (-10.0, 50.0)
    """
    lower = np.floor(yrange[0] / range_as_multiple_of) * range_as_multiple_of
    upper = np.ceil(yrange[1] / range_as_multiple_of) * range_as_multiple_of

    # Calculate the range needed to be divisible by both nticks-1 and range_as_multiple_of
    range_needed = np.lcm(nticks - 1, range_as_multiple_of)

    # Calculate how many complete range_needed units fit in our current range
    current_range = upper - lower
    num_units = int(np.ceil(current_range / range_needed))

    # Adjust the upper bound to make the range divisible by both factors
    upper = lower + num_units * range_needed

    return (float(lower), float(upper))


###############################################################################


if __name__ == "__main__":
    import doctest

    doctest.testmod()



# SPDX-FileCopyrightText: 2025 Florian Obersteiner / KIT
# SPDX-FileContributor: Florian Obersteiner <f.obersteiner@kit.edu>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import unittest
from pathlib import Path

from pyfuppes import txt2dict

try:
    wd = Path(__file__).parent
except NameError:
    wd = Path.cwd()
assert wd.is_dir(), "faild to obtain working directory"

src = wd / "test_input" / "validate_na"


class TestTimecorr(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # to run before all tests
        pass

    @classmethod
    def tearDownClass(cls):
        # to run after all tests
        pass

    def setUp(self):
        # to run before each test
        pass

    def tearDown(self):
        # to run after each test
        pass

    def test_txt2dict_basic(self):
        # this just reads each line until delimiter is encountered
        d = txt2dict.txt_2_dict_basic(
            src / "OM_20200304_591_CPT_MUC_V01_valid1.txt", offset=0, delimiter=" "
        )
        self.assertListEqual(["37"], list(d.keys()))

        # correct offset and delimiter:
        d = txt2dict.txt_2_dict_basic(
            src / "OM_20200304_591_CPT_MUC_V01_valid1.txt", offset=36, delimiter="\t"
        )
        self.assertListEqual(["TimeCRef", "Özone"], list(d.keys()))

    def test_txt2dict(self):
        # invalid settings
        with self.assertRaises(ValueError):
            _ = txt2dict.txt_2_dict(src / "OM_20200304_591_CPT_MUC_V01_valid1.txt")

        d = txt2dict.txt_2_dict(
            src / "OM_20200304_591_CPT_MUC_V01_valid1.txt", colhdr_ix=36, delimiter="\t"
        )
        self.assertListEqual(["TimeCRef", "Özone"], list(d.data.keys()))
        self.assertEqual(len(d.file_hdr), 36)


if __name__ == "__main__":
    unittest.main()



# SPDX-FileCopyrightText: 2025 Florian Obersteiner / KIT
# SPDX-FileContributor: Florian Obersteiner <f.obersteiner@kit.edu>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import unittest

import numpy as np

from pyfuppes import avgbinmap


class TestAvgbinmap(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # to run before all tests
        pass

    @classmethod
    def tearDownClass(cls):
        # to run after all tests
        pass

    def setUp(self):
        # to run before each test
        pass

    def tearDown(self):
        # to run after each test
        pass

    def test_mean_angle(self):
        for angles, mean in zip(
            [[350, 10], [90, 180, 270, 360], [10, 20, 30]], (0, -90, 20), strict=False
        ):
            self.assertAlmostEqual(avgbinmap.mean_angle(np.array(angles)), mean)
            self.assertAlmostEqual(avgbinmap.mean_angle_sc(np.array(angles)), mean)
            self.assertAlmostEqual(avgbinmap.mean_angle_numba(np.array(angles)), mean)

    def test_mean_dayfrac(self):
        for angles, mean in zip(
            [[350, 10], [90, 180, 270, 360], [10, 20, 30]], (0, 270, 20), strict=False
        ):
            self.assertAlmostEqual(
                avgbinmap.mean_day_frac(np.array(angles) / 360, use_numba=False),
                mean / 360,
            )
            self.assertAlmostEqual(
                avgbinmap.mean_day_frac(np.array(angles) / 360, use_numba=True),
                mean / 360,
            )

    def test_get_npnanmean(self):
        """Test the get_npnanmean function."""
        # Test with array with no NaN values
        test_array = np.array([1.0, 2.0, 3.0, 4.0, 5.0])
        result = avgbinmap.get_npnanmean(test_array)
        expected = 3.0
        self.assertTrue(np.isclose(result, expected), f"Expected {expected}, got {result}")

        # Test with array containing NaN values
        test_array_with_nan = np.array([1.0, 2.0, np.nan, 4.0, 5.0])
        result = avgbinmap.get_npnanmean(test_array_with_nan)
        expected = 3.0
        self.assertTrue(np.isclose(result, expected), f"Expected {expected}, got {result}")

        # Test with all NaN values
        test_array_all_nan = np.array([np.nan, np.nan, np.nan])
        result = avgbinmap.get_npnanmean(test_array_all_nan)
        self.assertTrue(np.isnan(result), f"Expected NaN, got {result}")

        # Test with empty array
        test_empty_array = np.array([])
        result = avgbinmap.get_npnanmean(test_empty_array)
        self.assertTrue(np.isnan(result), f"Expected NaN, got {result}")

    def test_bin_t_10s(self):
        t = np.arange(10) - 1  # -1 excludes first value
        d = avgbinmap.bin_t_10s(t)
        self.assertEqual(d.t_binned, np.array([5.0]))
        self.assertEqual(d.masked_bins.size, 0)
        self.assertTrue(
            (
                d.masked_vals
                == np.array([True, False, False, False, False, False, False, False, False, False])
            ).all()
        )

    def test_bin_y_of_t(self):
        t = np.arange(10) - 1  # -1 excludes first value
        v = np.arange(10)
        d = avgbinmap.bin_t_10s(t)
        want = np.array([np.average(v[1:])])
        have = avgbinmap.bin_y_of_t(v, d, aggregation="mean")
        self.assertListEqual(list(want), list(have))

    def test_bin_by_pdresample(self):
        t = np.arange(10) - 1  # -1 excludes first value
        v = np.arange(10) + 55
        df = avgbinmap.bin_by_pdresample(t, v)
        want = np.array([np.average(v[1:])])
        have = df["v_0"].to_list()
        self.assertListEqual(want.tolist(), have)
        want = np.array([5.0])
        have = df.index.to_list()
        self.assertListEqual(want.tolist(), have)

    def test_bin_by_npreduceat(self):
        v = np.arange(10) + 45
        have = avgbinmap.bin_by_npreduceat(v, 1)
        want = np.array([np.average(v)])
        self.assertEqual(want, have)
        have = avgbinmap.bin_by_npreduceat(v, 2)
        want = np.array([np.average(v[:5]), np.average(v[5:])])
        self.assertTrue((want == have).all())

    def test_mvg_avg(self):
        v = np.random.rand(100)
        v[40:] += 2
        v[70:] -= 1
        N = 5
        avg_np = avgbinmap.np_mvg_avg(v, N)
        avg_pd = avgbinmap.pd_mvg_avg(v, N)
        avg_sp = avgbinmap.sp_mvg_avg(v, N)
        self.assertTrue(np.isclose(avg_np[N - 1 : -N + 1], avg_sp[N - 1 : -N + 1]).all())
        self.assertTrue(np.isclose(avg_np[N - 1 : -N + 1], avg_pd[N - 1 : -N + 1]).all())

    def test_map_dependent(self):
        # first missing
        xref = np.array([1, 2, 3], dtype=int)
        xcmp = np.array([2, 3, 4], dtype=int)
        vcmp = np.array([1, 2, 3], dtype=float)
        tgt = np.array([np.nan, 1, 2], dtype=float)
        test = avgbinmap.map_dependent(xref, xcmp, vcmp)
        self.assertTrue(
            all(a == b if np.isfinite(b) else np.isnan(a) for a, b in zip(test, tgt, strict=False))
        )

        # last missing
        xref = np.array([1, 2, 3], dtype=int)
        xcmp = np.array([0, 1, 2], dtype=int)
        vcmp = np.array([1, 2, 3], dtype=float)
        tgt = np.array([2, 3, np.nan], dtype=float)
        test = avgbinmap.map_dependent(xref, xcmp, vcmp)
        self.assertTrue(
            all(a == b if np.isfinite(b) else np.isnan(a) for a, b in zip(test, tgt, strict=False))
        )

        # gap
        xref = np.array([1, 2, 3, 4], dtype=int)
        xcmp = np.array([1, 4, 5, 6], dtype=int)
        vcmp = np.array([1, 2, 3, 4], dtype=float)
        tgt = np.array([1, np.nan, np.nan, 2], dtype=float)
        test = avgbinmap.map_dependent(xref, xcmp, vcmp)
        self.assertTrue(
            all(a == b if np.isfinite(b) else np.isnan(a) for a, b in zip(test, tgt, strict=False))
        )

        # missing elements in cmp
        xref = np.array([1, 2, 3, 4], dtype=int)
        xcmp = np.array([1, 4, 5], dtype=int)
        vcmp = np.array([1, 2, 3], dtype=float)
        tgt = np.array([1, np.nan, np.nan, 2], dtype=float)
        test = avgbinmap.map_dependent(xref, xcmp, vcmp)
        self.assertTrue(
            all(a == b if np.isfinite(b) else np.isnan(a) for a, b in zip(test, tgt, strict=False))
        )

        # missing elements in ref
        xref = np.array([1, 2, 3], dtype=int)
        xcmp = np.array([1, 4, 5, 6], dtype=int)
        vcmp = np.array([1, 2, 3, 4], dtype=float)
        tgt = np.array([1, np.nan, np.nan], dtype=float)
        test = avgbinmap.map_dependent(xref, xcmp, vcmp)
        self.assertTrue(
            all(a == b if np.isfinite(b) else np.isnan(a) for a, b in zip(test, tgt, strict=False))
        )

    def test_calc_shift(self):
        have = np.array([1.2, 2.1, 2.5, 4.1, 5.3, 6.0, 7.1, 7.9, 9.0, 10.6], dtype=float)
        want = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10], dtype=float)
        shift = avgbinmap.calc_shift(have, step=1, lower_bound=-2, upper_bound=3)
        self.assertTrue(np.isclose(have + shift, want).all())

        have = np.array([0.1, 1.2, 2.3, 3.1], dtype=float)
        want = np.array([0, 1, 2, 3], dtype=float)
        shift = avgbinmap.calc_shift(have, step=1.0)
        self.assertTrue(np.isclose(have + shift, want).all())


if __name__ == "__main__":
    # import matplotlib.pyplot as plt
    #
    # v = np.random.rand(100)
    # v[40:] += 2
    # v[70:] -= 1
    # N = 5
    # avg_np = avgbinmap.np_mvg_avg(v, N)
    # avg_pd = avgbinmap.pd_mvg_avg(v, N)
    # avg_sp = avgbinmap.sp_mvg_avg(v, N)
    #
    # plt.plot(v, label="input")
    # plt.plot(avg_pd, label="np")
    # plt.plot(avg_pd, label="pd")
    # plt.plot(avg_sp, label="sp")
    # plt.legend()
    # plt.grid()
    # plt.show()
    #
    unittest.main()



# SPDX-FileCopyrightText: 2025 Florian Obersteiner / KIT
# SPDX-FileContributor: Florian Obersteiner <f.obersteiner@kit.edu>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import unittest

import numpy as np
import polars as pl
from polars.testing import assert_frame_not_equal

from pyfuppes import timecorr


def _make_df(dt_list):
    df = pl.DataFrame({"datetime": dt_list, "values": list(range(len(dt_list)))})
    return df.with_columns(
        pl.col("datetime").str.strptime(pl.Datetime, format="%Y-%m-%d").cast(pl.Datetime)
    )


class TestTimecorr(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # to run before all tests
        pass

    @classmethod
    def tearDownClass(cls):
        # to run after all tests
        pass

    def setUp(self):
        # to run before each test
        pass

    def tearDown(self):
        # to run after each test
        pass

    def test_correct_time(self):
        # offset only
        order = 1
        t = np.array([1, 2, 3, 4, 5, 6], dtype=np.float32)
        r = np.array([2, 3, 4, 5, 6, 7], dtype=np.float32)
        result = timecorr.correct_time(t, r, order)
        self.assertTrue(np.isclose(result.t_corr, r).all())

        t = np.array([-2, -1, 0, 1, 2, 3], dtype=np.float32)
        r = np.array([1, 2, 3, 4, 5, 6], dtype=np.float32)
        result = timecorr.correct_time(t, r, order)
        self.assertTrue(np.isclose(result.t_corr, r).all())

        # changed order
        order = 3
        t = np.array([1, 2, 3, 4, 5, 6], dtype=np.float32)
        r = np.array([2, 3, 4, 5, 6, 7], dtype=np.float32)
        result = timecorr.correct_time(t, r, order)
        self.assertTrue(np.isclose(result.t_corr, r).all())

        # inclination only
        order = 1
        t = np.array([1, 2, 3, 4, 5, 6], dtype=np.float32)
        r = np.array([1, 3, 5, 7, 9, 11], dtype=np.float32)
        result = timecorr.correct_time(t, r, order)
        self.assertTrue(np.isclose(result.t_corr, r).all())

        # inclination + offset
        order = 1
        t = np.array([1, 2, 3, 4, 5, 6], dtype=np.float32)
        r = np.array([2.0, 3.5, 5.0, 6.5, 8.0, 9.5], dtype=np.float32)
        result = timecorr.correct_time(t, r, order)
        self.assertTrue(np.isclose(result.t_corr, r).all())

    def test_pldt_filter(self):
        # edge case: first invalid
        have = ["2022-10-30", "2022-10-28", "2022-10-29"]
        # want_forward = ["2022-10-30"]
        # want_backward = ["2022-10-28", "2022-10-29"]

        df = _make_df(have)
        n, df_out = timecorr.filter_dt_forward(df)
        self.assertEqual(n, 2)
        self.assertTrue((df_out["values"] == pl.Series([0])).all())
        assert_frame_not_equal(df, df_out)

        df = _make_df(have)
        n, df_out = timecorr.filter_dt_backward(df)
        self.assertEqual(n, 1)
        self.assertTrue((df_out["values"] == pl.Series([1, 2])).all())
        assert_frame_not_equal(df, df_out)

        # edge case: last invalid
        have = ["2022-10-29", "2022-10-30", "2022-10-29"]
        # want_forward = ["2022-10-29", "2022-10-30"]

        df = _make_df(have)
        n, df = timecorr.filter_dt_forward(df)
        self.assertEqual(n, 1)
        self.assertTrue((df["values"] == pl.Series([0, 1])).all())

        # some invalid key
        have = ["2022-10-29", "2022-10-30", "2022-10-28", "2022-10-31"]
        # want_forward = ["2022-10-29", "2022-10-30", "2022-10-31"]
        # want_backward = ["2022-10-28", "2022-10-31"]

        df = _make_df(have)
        n, df = timecorr.filter_dt_forward(df)
        self.assertEqual(n, 1)
        self.assertTrue((df["values"] == pl.Series([0, 1, 3])).all())

        df = _make_df(have)
        n, df = timecorr.filter_dt_backward(df)
        self.assertEqual(n, 2)
        self.assertTrue((df["values"] == pl.Series([2, 3])).all())

        have = ["2022-10-29", "2022-10-30", "2022-11-01", "2022-10-31"]
        # want_forward = ["2022-10-29", "2022-10-30", "2022-11-01"]
        # want_backward = ["2022-10-29", "2022-10-30", "2022-10-31"]

        df = _make_df(have)
        n, df = timecorr.filter_dt_forward(df)
        self.assertEqual(n, 1)
        self.assertTrue((df["values"] == pl.Series([0, 1, 2])).all())

        df = _make_df(have)
        n, df = timecorr.filter_dt_backward(df)
        self.assertEqual(n, 1)
        self.assertTrue((df["values"] == pl.Series([0, 1, 3])).all())

        have = ["2022-10-29", "2022-10-30", "2022-10-28", "2022-10-30"]
        # want_forward = ["2022-10-29", "2022-10-30"]
        # want_backward = ["2022-10-28", "2022-10-30"]

        df = _make_df(have)
        n, df = timecorr.filter_dt_forward(df)
        self.assertEqual(n, 2)
        self.assertTrue((df["values"] == pl.Series([0, 1])).all())

        df = _make_df(have)
        n, df_out = timecorr.filter_dt_backward(df)
        self.assertEqual(n, 2)
        self.assertTrue((df_out["values"] == pl.Series([2, 3])).all())
        assert_frame_not_equal(df, df_out)

    def test_xcorr_timelag(self):
        # signal with peak
        t = np.linspace(0, 250, 250)
        f = 10 * np.exp(-((t - 90) ** 2) / 8) + np.random.randn(250) + 99
        g = 10 * np.exp(-((t - 180) ** 2) / 8) + np.random.randn(250) + 41
        lag = timecorr.xcorr_timelag(t, f, t, g, show_plots=False)
        # print(l, 90)
        self.assertTrue(abs(lag - 90) < (90 * 0.02))  # shift is 90... expect within 2%

        # Sawtooth wave
        # https://stackoverflow.com/questions/4688715/find-time-shift-between-two-similar-waveforms
        f = np.array([0, 1, 2, 3, 4, 3, 2, 1, 0, 1, 2, 3, 4, 3, 2, 1, 0, 0, 0, 0, 0])
        g = np.array([0, 0, 0, 0, 0, 1, 2, 3, 4, 3, 2, 1, 0, 1, 2, 3, 4, 3, 2, 1, 0])
        t = np.arange(f.size)
        lag = timecorr.xcorr_timelag(t, f, t, g, show_plots=False)
        # print(l, 4)
        # shift is 4... expect within 10%
        self.assertTrue(abs(lag - 4) < (4 * 0.1))

        # sine waves with noise
        # https://stackoverflow.com/questions/41492882/find-time-shift-of-two-signals-using-cross-correlation
        to_freq = 50
        data = np.linspace(0, 2 * np.pi, to_freq)
        data = np.tile(np.sin(data), 5)
        data += np.random.normal(0, 5, data.shape)
        f, g = data[to_freq : 4 * to_freq], data[: 3 * to_freq]
        t = np.arange(f.size)
        lag = timecorr.xcorr_timelag(t, f, t, g, show_plots=False)
        # print(l, to_freq)
        self.assertTrue(
            abs(lag - to_freq) < (to_freq * 0.02)
        )  # shift is to_freq... expect within 2%

        # more waves
        # https://stackoverflow.com/questions/13826290/estimating-small-time-shift-between-two-time-series
        shift = -3
        x0 = np.linspace(0, 10, 100)
        x1 = 1 * x0 + shift
        f = sum([np.sin(2 * np.pi * i * x0 / 10) for i in range(1, 5)])
        g = sum([np.sin(2 * np.pi * i * x1 / 10) for i in range(1, 5)])
        lag = timecorr.xcorr_timelag(x0, f, x0, g, xrange=(x0.min(), x1.max()), show_plots=False)  # type: ignore
        # print(l, shift*-1)
        # shift is 90... expect within 2%
        self.assertTrue(abs(lag - 3) < (3 * 0.02))


if __name__ == "__main__":
    unittest.main()



# SPDX-FileCopyrightText: 2025 Florian Obersteiner / KIT
# SPDX-FileContributor: Florian Obersteiner <f.obersteiner@kit.edu>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import unittest

import numpy as np
import pandas as pd
import polars as pl
from pandas import testing as pdtest
from polars.testing import assert_series_equal

from pyfuppes import interpolate


class TestInterpolate(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        pass

    @classmethod
    def tearDownClass(cls):
        # to run after all tests
        pass

    def setUp(self):
        # to run before each test
        pass

    def tearDown(self):
        # to run after each test
        pass

    # --------------------------------------------------------------------------------

    def test_interp_df(self):
        # high to low freq:
        df_src = pd.DataFrame(
            {
                "dt": [
                    "2022-12-14T14:00:01.000",
                    "2022-12-14T14:00:02.000",
                    "2022-12-14T14:00:03.000",
                    "2022-12-14T14:00:04.000",
                    "2022-12-14T14:00:05.000",
                    "2022-12-14T14:00:06.000",
                ],
                "v1": [1.0, 2.0, 3.0, 4.0, 5.0, 6.0],
            }
        )
        df_src.index = pd.to_datetime(df_src["dt"]).astype("int64")
        new_index = (
            pd.to_datetime(
                [
                    "2022-12-14T14:00:01.500",
                    "2022-12-14T14:00:03.500",
                    "2022-12-14T14:00:05.500",
                ]
            )
            .to_series()
            .astype("int64")
        )
        new_index.name = df_src.index.name
        df_out = interpolate.pd_DataFrame_ip(df_src, new_index)
        self.assertIsNone(
            pdtest.assert_series_equal(
                df_out["v1_ip"], pd.Series([1.5, 3.5, 5.5], name="v1_ip", index=new_index)
            )
        )

        # low to high freq:
        df_src = pd.DataFrame(
            {
                "dt": [
                    "2022-12-14T14:00:01.500",
                    "2022-12-14T14:00:03.500",
                    "2022-12-14T14:00:05.500",
                ],
                "v1": [1.5, 3.5, 5.5],
            }
        )
        df_src.index = pd.to_datetime(df_src["dt"]).astype("int64")
        new_index = (
            pd.to_datetime(
                [
                    "2022-12-14T14:00:01.000",
                    "2022-12-14T14:00:02.000",
                    "2022-12-14T14:00:03.000",
                    "2022-12-14T14:00:04.000",
                    "2022-12-14T14:00:05.000",
                    "2022-12-14T14:00:06.000",
                ]
            )
            .to_series()
            .astype("int64")
        )
        new_index.name = df_src.index.name
        df_out = interpolate.pd_DataFrame_ip(df_src, new_index)
        self.assertIsNone(
            pdtest.assert_series_equal(
                df_out["v1_ip"],
                # edge values are only repeated, does not extrapolate:
                pd.Series([1.5, 2.0, 3.0, 4.0, 5.0, 5.5], name="v1_ip", index=new_index),
            )
        )

    # --------------------------------------------------------------------------------

    def test_pd_Series_interp1d(self):
        # low to high freq:
        df_dst = pd.DataFrame(
            {
                "dt": [
                    "2022-12-14T14:00:01.000",
                    "2022-12-14T14:00:02.000",
                    "2022-12-14T14:00:03.000",
                    "2022-12-14T14:00:04.000",
                    "2022-12-14T14:00:05.000",
                    "2022-12-14T14:00:06.000",
                ]
            }
        )
        df_dst["dt"] = pd.to_datetime(df_dst["dt"]).astype("int64")
        df_src = pd.DataFrame(
            {
                "dt": [
                    "2022-12-14T14:00:01.500",
                    "2022-12-14T14:00:03.500",
                    "2022-12-14T14:00:05.500",
                ],
                "v1": [1.5, 3.5, 5.5],
            }
        )
        df_src["dt"] = pd.to_datetime(df_src["dt"]).astype("int64")
        df_out = interpolate.pd_Series_ip1d(
            df_src,
            df_dst,
            ivar_src_name="dt",
            dvar_src_name="v1",
            ivar_dst_name="dt",
            dvar_dst_name="v1_ip",
            kind="linear",  # no effect with pl_Series_ip1d_lite
            bounds_error=False,  # no effect with pl_Series_ip1d_lite
            fill_value="extrapolate",
        )
        with self.assertRaises(AssertionError):
            pdtest.assert_frame_equal(df_dst, df_out)

        self.assertIsNone(
            pdtest.assert_series_equal(
                df_out["v1_ip"], pd.Series([1.0, 2.0, 3.0, 4.0, 5.0, 6.0], name="v1_ip")
            )
        )

        df_out = interpolate.pd_Series_ip1d(
            df_src,
            df_dst,
            ivar_src_name="dt",
            dvar_src_name="v1",
            ivar_dst_name="dt",
            dvar_dst_name="v1_ip",
            kind="linear",  # no effect with pl_Series_ip1d_lite
            bounds_error=False,  # no effect with pl_Series_ip1d_lite
            fill_value=np.nan,
        )
        self.assertIsNone(
            pdtest.assert_series_equal(
                df_out["v1_ip"],
                pd.Series([np.nan, 2.0, 3.0, 4.0, 5.0, np.nan], name="v1_ip"),
            )
        )
        # high to low freq:
        df_src = pd.DataFrame(
            {
                "dt": [
                    "2022-12-14T14:00:01.000",
                    "2022-12-14T14:00:02.000",
                    "2022-12-14T14:00:03.000",
                    "2022-12-14T14:00:04.000",
                    "2022-12-14T14:00:05.000",
                    "2022-12-14T14:00:06.000",
                ],
                "v1": [1.0, 2.0, 3.0, 4.0, 5.0, 6.0],
            }
        )
        df_src["dt"] = pd.to_datetime(df_src["dt"]).astype("int64")
        df_dst = pd.DataFrame(
            {
                "dt": [
                    "2022-12-14T14:00:01.500",
                    "2022-12-14T14:00:03.500",
                    "2022-12-14T14:00:05.500",
                ],
            }
        )
        df_dst["dt"] = pd.to_datetime(df_dst["dt"]).astype("int64")
        df_out = interpolate.pd_Series_ip1d(
            df_src,
            df_dst,
            ivar_src_name="dt",
            ivar_dst_name="dt",
            dvar_src_name="v1",
            dvar_dst_name="v1_ip",
            kind="linear",  # no effect with pl_Series_ip1d_lite
            bounds_error=True,  # no effect with pl_Series_ip1d_lite
        )
        with self.assertRaises(AssertionError):
            pdtest.assert_frame_equal(df_dst, df_out)
        self.assertIsNone(
            pdtest.assert_series_equal(df_out["v1_ip"], pd.Series([1.5, 3.5, 5.5], name="v1_ip"))
        )

    # --------------------------------------------------------------------------------

    def test_pl_Series_interp1d(self):
        for func in interpolate.pl_Series_interp1d, interpolate.pl_Series_ip1d_lite:
            # low to high freq:
            df_dst = pl.DataFrame(
                {
                    "dt": [
                        "2022-12-14T14:00:01.000",
                        "2022-12-14T14:00:02.000",
                        "2022-12-14T14:00:03.000",
                        "2022-12-14T14:00:04.000",
                        "2022-12-14T14:00:05.000",
                        "2022-12-14T14:00:06.000",
                    ]
                }
            )
            df_dst = df_dst.with_columns(pl.col("dt").str.strptime(pl.Datetime))
            df_src = pl.DataFrame(
                {
                    "dt": [
                        "2022-12-14T14:00:01.500",
                        "2022-12-14T14:00:03.500",
                        "2022-12-14T14:00:05.500",
                    ],
                    "v1": [1.5, 3.5, 5.5],
                }
            )
            df_src = df_src.with_columns(pl.col("dt").str.strptime(pl.Datetime))
            df_dst = func(
                df_src,
                df_dst,
                ivar_src_name="dt",
                ivar_dst_name="dt",
                dvar_src_name="v1",
                dvar_dst_name="v1_ip",
                kind="linear",  # no effect with pl_Series_ip1d_lite
                bounds_error=False,  # no effect with pl_Series_ip1d_lite
                fill_value="extrapolate",
            )
            self.assertIsNone(
                assert_series_equal(
                    df_dst["v1_ip"], pl.Series("v1_ip", [1.0, 2.0, 3.0, 4.0, 5.0, 6.0])
                )
            )
            df_dst = func(
                df_src,
                df_dst,
                ivar_src_name="dt",
                ivar_dst_name="dt",
                dvar_src_name="v1",
                dvar_dst_name="v1_ip",
                kind="linear",  # no effect with pl_Series_ip1d_lite
                bounds_error=False,  # no effect with pl_Series_ip1d_lite
                fill_value=np.nan,
            )
            self.assertIsNone(
                assert_series_equal(
                    df_dst["v1_ip"],
                    pl.Series("v1_ip", [np.nan, 2.0, 3.0, 4.0, 5.0, np.nan]),
                )
            )
            # high to low freq:
            df_src = pl.DataFrame(
                {
                    "dt": [
                        "2022-12-14T14:00:01.000",
                        "2022-12-14T14:00:02.000",
                        "2022-12-14T14:00:03.000",
                        "2022-12-14T14:00:04.000",
                        "2022-12-14T14:00:05.000",
                        "2022-12-14T14:00:06.000",
                    ],
                    "v1": [1.0, 2.0, 3.0, 4.0, 5.0, 6.0],
                }
            )
            df_src = df_src.with_columns(pl.col("dt").str.strptime(pl.Datetime))
            df_dst = pl.DataFrame(
                {
                    "dt": [
                        "2022-12-14T14:00:01.500",
                        "2022-12-14T14:00:03.500",
                        "2022-12-14T14:00:05.500",
                    ],
                }
            )
            df_dst = df_dst.with_columns(pl.col("dt").str.strptime(pl.Datetime))
            df_dst = func(
                df_src,
                df_dst,
                ivar_src_name="dt",
                ivar_dst_name="dt",
                dvar_src_name="v1",
                dvar_dst_name="v1_ip",
                kind="linear",  # no effect with pl_Series_ip1d_lite
                bounds_error=True,  # no effect with pl_Series_ip1d_lite
            )
            self.assertIsNone(
                assert_series_equal(df_dst["v1_ip"], pl.Series("v1_ip", [1.5, 3.5, 5.5]))
            )


if __name__ == "__main__":
    import warnings

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=UserWarning)
        unittest.main()



# SPDX-FileCopyrightText: 2025 Florian Obersteiner / KIT
# SPDX-FileContributor: Florian Obersteiner <f.obersteiner@kit.edu>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import unittest

import numpy as np

from pyfuppes import filters


class TestFilters(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # to run before all tests
        pass

    @classmethod
    def tearDownClass(cls):
        # to run after all tests
        pass

    def setUp(self):
        # to run before each test
        pass

    def tearDown(self):
        # to run after each test
        pass

    def test_mask_repeated(self):
        arr = np.array([1, 2, 3, 3, 3, 4, 5, 5])
        #                           |-> nok
        have = filters.mask_repeated(arr, 2)
        want = np.array([True, True, True, True, False, True, True, True])
        self.assertTrue((have == want).all())

    def test_mask_repeated_nb(self):
        arr = np.array([1, 2, 3, 3, 3, 4, 5, 5])
        #                           |-> nok
        have = filters.mask_repeated_nb(arr, 2)
        want = np.array([True, True, True, True, False, True, True, True])
        self.assertTrue((have == want).all())

    def test_mask_jumps(self):
        arr = np.array([1, 2, 3, 7, 3, 4, 5, 5])
        #                        |-> nok
        have = filters.mask_jumps(arr, 3, 1, use_abs_delta=False)
        want = np.array([True, True, True, False, True, True, True, True])
        self.assertTrue((have == want).all())
        # changing lookahead must not change result here:
        have = filters.mask_jumps(arr, 3, 2, use_abs_delta=False)
        want = np.array([True, True, True, False, True, True, True, True])
        self.assertTrue((have == want).all())

        arr = np.array([1, 2, 3, -7, 3, 4, 5, 5])
        #                            |-> nok
        have = filters.mask_jumps(arr, 3, 1, use_abs_delta=False)
        want = np.array([True, True, True, True, False, True, True, True])
        self.assertTrue((have == want).all())

        arr = np.array([1, 2, 3, -7, 3, 4, 5, 5])
        #                         |-> nok
        have = filters.mask_jumps(arr, 3, 1, use_abs_delta=True)
        want = np.array([True, True, True, False, True, True, True, True])
        self.assertTrue((have == want).all())

        arr = np.array([1, 2, 3, 8, 7, 4, 5, 5])
        #                        |--|-> nok
        have = filters.mask_jumps(arr, 3, 2, use_abs_delta=False)
        want = np.array([True, True, True, False, False, True, True, True])
        self.assertTrue((have == want).all())

        arr = np.array([1, 2, 3, -8, 7, 4, 5, 5])
        #                        |--|-> nok
        have = filters.mask_jumps(arr, 3, 2, use_abs_delta=True)
        want = np.array([True, True, True, False, False, True, True, True])
        self.assertTrue((have == want).all())

        arr = np.array([1, 2, 3, 8, -7, 4, 5, 5])
        #                        |--|-> nok
        have = filters.mask_jumps(arr, 3, 2, use_abs_delta=True)
        want = np.array([True, True, True, False, False, True, True, True])
        self.assertTrue((have == want).all())

    def test_filter_jumps(self):
        arr = np.array([1, 2, 3, 7, 3, 4, 5, 5], dtype=float)
        #                        |-> nok
        have0 = filters.filter_jumps(arr, 3, 1, use_abs_delta=False)
        have1 = filters.filter_jumps_np(arr, 3, 1, use_abs_delta=False)
        want = arr[np.array([True, True, True, False, True, True, True, True])]
        self.assertTrue((have0[0][have0[1]] == want).all())
        self.assertTrue((have1.filtered == want).all())

        arr = np.array([1, 2, 3, 8, 7, 4, 5, 5], dtype=float)
        #                        |--|-> nok
        have0 = filters.filter_jumps(arr, 3, 2, use_abs_delta=False)
        have1 = filters.filter_jumps_np(arr, 3, 2, use_abs_delta=False)
        want = arr[np.array([True, True, True, False, False, True, True, True])]
        self.assertTrue((have0[0][have0[1]] == want).all())
        self.assertTrue((have1.filtered == want).all())

        arr = np.array([1, 2, 3, -8, 7, 4, 5, 5], dtype=float)
        #                        |--|-> nok
        have0 = filters.filter_jumps(arr, 3, 2, use_abs_delta=True)
        have1 = filters.filter_jumps_np(arr, 3, 2, use_abs_delta=True)
        want = arr[np.array([True, True, True, False, False, True, True, True])]
        self.assertTrue((have0[0][have0[1]] == want).all())
        self.assertTrue((have1.filtered == want).all())

        arr = np.array([1, 2, 3, 8, -7, 4, 5, 5], dtype=float)
        #                        |--|-> nok
        have0 = filters.filter_jumps(arr, 3, 2, use_abs_delta=True)
        have1 = filters.filter_jumps_np(arr, 3, 2, use_abs_delta=True)
        want = arr[np.array([True, True, True, False, False, True, True, True])]
        self.assertTrue((have0[0][have0[1]] == want).all())
        self.assertTrue((have1.filtered == want).all())

    def test_extend_mask(self):
        m = np.array([False, False, True, False, False])
        have = filters.extend_mask(m, 0)  # n = 0 has no effect
        want = np.array([False, False, True, False, False])
        self.assertTrue((have == want).all())

        have = filters.extend_mask(m, 1)  # one 'True' inserted to the right
        want = np.array([False, False, True, True, False])
        self.assertTrue((have == want).all())

        have = filters.extend_mask(m, 2)  # another 'True' inserted to the left
        want = np.array([False, True, True, True, False])
        self.assertTrue((have == want).all())

        # edge
        m = np.array([False, False, False, False, True])
        have = filters.extend_mask(m, 1)  # n = 1 has no effect here, no element to the right
        want = np.array([False, False, False, False, True])
        self.assertTrue((have == want).all())

        have = filters.extend_mask(m, 2)  # n = 2 gives one 'True' to the left
        want = np.array([False, False, False, True, True])
        self.assertTrue((have == want).all())

        have = filters.extend_mask(m, 3)  # n = 3 gives same result as n = 2
        want = np.array([False, False, False, True, True])
        self.assertTrue((have == want).all())

        m = np.array(
            [
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                True,
                False,
                False,
                False,
                False,
                False,
                False,
                True,
                False,
                False,
            ]
        )
        have = filters.extend_mask(m, 5)
        want = np.array(
            [
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                True,
                True,
                True,
                True,
                True,
                True,
                False,
                True,
                True,
                True,
                True,
                True,
            ]
        )
        self.assertTrue((have == want).all())

    def test_filter_lof(self):
        with self.assertRaises(Exception) as context:
            _ = filters.simple_1d_lof(np.array([np.nan]), 15, 1.5)
        self.assertTrue("can only use all-finite values in 'var'" in str(context.exception))

        with self.assertRaises(Exception) as context:
            _ = filters.simple_1d_lof(np.array([[1, 1], [2, 2]]), 15, 1.5)
        self.assertTrue("can only work with 1D data" in str(context.exception))

        np.random.seed(42)
        rng = np.random.default_rng()
        data = rng.standard_normal(10000)
        n_outliers = 20

        m = np.sort(rng.choice(data.shape[0], size=n_outliers, replace=False))
        pos_neg = np.random.choice([True, False], size=n_outliers)

        out = np.absolute(np.min(data) - np.max(data)) * 5
        data[m[pos_neg]] = out
        data[m[~pos_neg]] = out * -1

        mask = filters.simple_1d_lof(data, 15, 1.5)
        self.assertEqual(
            m.shape[0], np.nonzero(mask)[0].shape[0], "lof must find prescribed outliers"
        )
        self.assertTrue((m == np.nonzero(mask)[0]).all(), "lof must find prescribed outliers")

        mask = filters.simple_1d_lof(data, 15, 1.5, mode="positive")
        self.assertEqual(
            m[pos_neg].shape[0], np.nonzero(mask)[0].shape[0], "lof must find prescribed outliers"
        )
        self.assertTrue(
            (m[pos_neg] == np.nonzero(mask)[0]).all(), "lof must find prescribed outliers"
        )

        mask = filters.simple_1d_lof(data, 15, 1.5, mode="negative")
        self.assertEqual(
            m[~pos_neg].shape[0], np.nonzero(mask)[0].shape[0], "lof must find prescribed outliers"
        )
        self.assertTrue(
            (m[~pos_neg] == np.nonzero(mask)[0]).all(), "lof must find prescribed outliers"
        )


if __name__ == "__main__":
    unittest.main()



# SPDX-FileCopyrightText: 2025 Florian Obersteiner / KIT
# SPDX-FileContributor: Florian Obersteiner <f.obersteiner@kit.edu>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import unittest
from datetime import UTC, datetime

import numpy as np

from pyfuppes import geo


class TestGeo(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # to run before all tests
        pass

    @classmethod
    def tearDownClass(cls):
        # to run after all tests
        pass

    def setUp(self):
        # to run before each test
        pass

    def tearDown(self):
        # to run after each test
        pass

    def test_sza(self):
        a = geo.sza(datetime(1970, 1, 1, tzinfo=UTC), 42, 55)
        b = geo.sza_pysolar(datetime(1970, 1, 1, tzinfo=UTC), 42, 55)
        self.assertAlmostEqual(a, b, delta=0.05)

        a = geo.sza(datetime(2013, 9, 2, 10, 47, tzinfo=UTC), 10, 33)
        b = geo.sza_pysolar(datetime(2013, 9, 2, 10, 47, tzinfo=UTC), 10, 33)
        self.assertAlmostEqual(a, b, delta=0.05)

    def test_haversine_dist(self):
        dist = 2887
        tol_decimalplaces = 0
        self.assertAlmostEqual(
            geo.haversine_dist(np.array((36.12, 33.94)), np.array((-86.67, -118.40))),
            dist,
            tol_decimalplaces,
        )

    def test_geodist(self):
        lat = np.array([-71.312796, -81.695391])
        lon = np.array([41.49008, 41.499498])
        dist = geo.geodesic_dist(lat, lon)
        self.assertAlmostEqual(dist.sum(), 1159.0067, places=4)


if __name__ == "__main__":
    unittest.main()



# SPDX-FileCopyrightText: 2025 Florian Obersteiner / KIT
# SPDX-FileContributor: Florian Obersteiner <f.obersteiner@kit.edu>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""NASA Ames FFI 1001 text file format reader / writer."""

from datetime import UTC, datetime
from pathlib import Path

import numpy as np
import pandas as pd
import polars as pl

from .na1001_backend import nasa_ames_1001_rw as rw
from .na1001_backend import nasa_ames_1001_tools as tools

###############################################################################


class FFI1001:
    r"""
    A class to work with NASA Ames files with format index 1001.

    Parameters
    ----------
    file : str or pathlib.Path or file-like
        data source.
    sep : str, optional
        General delimiter. The default is " ".
    sep_data : str, optional
        Delimiter used exclusively in data block. The default is "\t".
    strip_lines : bool, optional
        Remove surrounding whitespaces from all lines before parsing.
        The default is True.
    auto_nncoml : bool, optional
        Automatically determine number of lines in comment blocks.
        The default is True.
    rmv_repeated_seps : bool, optional
        Remove repeated delimiters (e.g. double space). The default is False.
    vscale_vmiss_vertical : bool, optional
        VSCALE and VMISS parameters are arranged vertically over multiple
        lines (1 entry per line) instead of in one line each.
        The default is False.
    vmiss_to_None : bool, optional
        Set True if missing values should be replaced with None. The default is False.
    ensure_ascii : bool, optional
        Enforce ASCII-decoding of the input. The default is False.
    allow_emtpy_data : bool, optional
        Allow header-only input. The default is False.

    Returns
    -------
    FFI 1001 class instance.
    """

    def __init__(self, file=None, **kwargs) -> None:
        today = datetime.now(UTC).date()
        self.NLHEAD: int = 14  # minimum number of header lines is 14
        self.ONAME: str = "data origin"
        self.ORG: str = "organization"
        self.SNAME: str = "sampling description"
        self.MNAME: str = "mission name"
        self.IVOL: int = 1
        self.NVOL: int = 1
        self.DATE: tuple[int, int, int] = (1970, 1, 1)
        self.RDATE: tuple[int, int, int] = (today.year, today.month, today.day)
        self.DX: int = 0
        self.XNAME: str = "x name"
        self.NV: int = 1
        self.VSCAL: list[str] = ["1"]
        self.VMISS: list[str] = ["-9999"]
        self._VNAME: list[str] = ["v names"]
        self.NSCOML: int = 0
        self._SCOM: list[str] = ["special comments"]
        self.NNCOML: int = 0
        self._NCOM: list[str] = ["normal comments"]
        self._X: list[str] = [""]
        self.V: list[list[str]] = [[""]]
        self._SRC: str = "path to file"
        self._HEADER: str = "file header"

        if file is not None:
            self.__from_file(file, **kwargs)

    @property
    def FFI(self):
        """FFI is always 1001, as the class name indicates..."""
        return 1001

    @property
    def SCOM(self) -> list[str]:
        """Special comments block"""
        return self._SCOM

    @SCOM.setter
    def SCOM(self, value: list[str]):
        self._SCOM, self.NSCOML = value, len(value)
        self.NLHEAD = 14 + self.NSCOML + self.NNCOML + self.NV

    @property
    def NCOM(self) -> list[str]:
        """Normal comments block"""
        return self._NCOM

    @NCOM.setter
    def NCOM(self, value: list[str]):
        self._NCOM, self.NNCOML = value, len(value)
        self.NLHEAD = 14 + self.NSCOML + self.NNCOML + self.NV

    @property
    def VNAME(self) -> list[str]:
        """Variables name block"""
        return self._VNAME

    @VNAME.setter
    def VNAME(self, value: list[str]):
        self._VNAME, self.NV = value, len(value)
        self.NLHEAD = 14 + self.NSCOML + self.NNCOML

    @property
    def X(self) -> list[str]:
        """Independent variable"""
        return self._X

    @X.setter
    def X(self, xarr: list[str]):
        self._X = xarr
        # calculate dx as unique diffs in X,
        # unique with floats might fail, so add a round to 4 decimals:
        dx = np.unique(np.diff(np.array(xarr, dtype=float)).round(4))
        # let dx be 0 if there's more than one unique diff
        dx = dx[0] if dx.size == 1 else 0
        # use an integer if dx is close to its integer value, else float:
        dx = int(dx) if np.isclose(dx, int(dx)) else dx
        self.DX = dx

    @property
    def V(self) -> list[list[str],]:  # type:ignore
        """Dependent variable"""
        return self._V

    @V.setter
    def V(self, vlists: list[list[str],]):  # type:ignore
        # assert (
        #     len(vlists) == self.NV
        # ), f"try to set {len(vlists)} dependent variables, but VNAMES specify {self.NV}"
        self._V = vlists

    # ------------------------------------------------------------------------------
    def __from_file(self, file: str | Path, **kwargs):
        """Load NASA Ames 1001 from text file."""
        nadict = rw.na1001_cls_read(file, **kwargs)
        for k in rw.KEYS:
            setattr(self, k, nadict[k])

    # ------------------------------------------------------------------------------
    def to_file(self, file: str | Path, **kwargs):
        """
        Write NASA Ames 1001 file from populated ffi_1001 class.
        Output text is UTF-8 encoded, to allow for a more contemporary set of characters.

        Parameters
        ----------
        file : str or pathlib.Path
            filepath and -name of the destination.
        sep : str, optional
            General delimiter. The default is " ".
        sep_data : str, optional
            Delimiter to separate data columns. The default is "\t".
        overwrite : int, optional
            Set to 1 to overwrite existing files. The default is 0 (no overwrite).
        verbose : bool, optional
            Verbose print output if True. The default is False.

        Returns
        -------
        int
            0 -> failed, 1 -> successful write, 2 -> successful overwrite.

        """
        return rw.na1001_cls_write(file, self.__dict__, **kwargs)

    # ------------------------------------------------------------------------------
    def to_dict_nparray(self, **kwargs) -> dict[str, np.ndarray]:
        """
        Make dictionary of numpy 1D arrays from FFI1001 class instance.

        Parameters
        ----------
        sel_vnames : list of string, optional
            VNAMEs to be converted. The default is None.
        clean_vnames : boolean, optional
            remove whitespaces from variable names. The default is False.
        vname_delimiter : string, optional
            Where to split entries in VNAME. The default is ';'.
        split_idx : int, optional
            Which part of split result to use, see splitVname. The default is 0.
        xdtype : data type, optional
            Data type for independent variable X. The default is np.float.
        vdtype : data type, optional
            Data type for dependent variable(s). The default is np.float.
        vmiss : missing value identifier, optional
            The default is np.nan.

        Returns
        -------
        dict
            dictionary holding numpy arrays for variables from the na1001 class
            instance.

        """
        return tools.naDict_2_npndarr(self.__dict__, **kwargs)

    # ------------------------------------------------------------------------------
    def to_pddf(self, **kwargs) -> pd.DataFrame:
        """
        Make a pandas DataFrame from NA 1001 data (X and V).

        Parameters
        ----------
        sep_colhdr : str, optional
            separator used in column header (last line of NCOM). Default is tab.
        idx_colhdr : int, optional
            look for column header in NCOM at index idx_colhdr. Default is -1.
        clean_colnames : boolean, optional
            remove whitespaces from variable names. The default is False.
        dtype : numpy array data type, optional
            data type to use for conversion to DataFrame. Default is float.
        add_datetime_index: boolean, optional
            add a DateTime index to the df. The default is False.

        Returns
        -------
        pandas.DataFrame
            dataframe with a column for X and one for each parameter in V.

        """
        return tools.naDict_2_pddf(self.__dict__, **kwargs)

    # ------------------------------------------------------------------------------
    def to_poldf(self, **kwargs) -> pl.DataFrame:
        """
        Make a polars DataFrame from NA 1001 data (X and V).

        Parameters
        ----------
        sep_colhdr : str, optional
            separator used in column header (last line of NCOM). Default is tab.
        idx_colhdr : int, optional
            look for column header in NCOM at index idx_colhdr. Default is -1.
        clean_colnames : boolean, optional
            remove whitespaces from variable names. The default is False.
        add_datetime: boolean, optional
            add a DateTime column to the df. The default is False.
        nan_to_none: boolean, optional
            fill NaN values with polars' Null. The default is False.

        Returns
        -------
        polars.DataFrame
            dataframe with a column for X and one for each parameter in V.

        """
        return tools.naDict_2_poldf(self.__dict__, **kwargs)

    # ------------------------------------------------------------------------------
    def __repr__(self) -> str:
        s = f"NASA Ames {self.FFI}\n---\n"
        s += "".join([
            f"{k.strip('_')} : {self.__dict__[k]}\n"
            for k in (
                "_SRC",
                "NLHEAD",
                "ONAME",
                "ORG",
                "SNAME",
                "MNAME",
                "IVOL",
                "NVOL",
                "DATE",
                "RDATE",
                "DX",
                "XNAME",
                "NV",
                "VSCAL",
                "VMISS",
                "_VNAME",
                "NSCOML",
                "_SCOM",
                "NNCOML",
                "_NCOM",
            )
        ])
        return s

    def __str__(self) -> str:
        s = "NASA Ames 1001\n"
        s += f"SRC: {self._SRC}\n---\n"
        s += "\n".join(f"{k}: {getattr(self, k)}" for k in ["ONAME", "ORG", "SNAME"])
        s += "\n" + ", ".join(f"{k}: {getattr(self, k)}" for k in ["DATE", "RDATE"])
        return s



# SPDX-FileCopyrightText: 2025 Florian Obersteiner / KIT
# SPDX-FileContributor: Florian Obersteiner <f.obersteiner@kit.edu>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Interpolation of timeseries data."""

import numpy as np
import pandas as pd
import polars as pl
from scipy.interpolate import interp1d  # type:ignore


def pd_DataFrame_ip(df: pd.DataFrame, new_index: pd.Series) -> pd.DataFrame:
    """
    Generate a new DataFrame with all numeric columns interpolated to the new_index.

    Uses numpy's interp function, see
    <https://numpy.org/doc/stable/reference/generated/numpy.interp.html>

    (!) limitations:
      - indices must be monotonically increasing
      - cannot extrapolate; edge values are simply repeated

    Parameters
    ----------
    df : pd.DataFrame
        the dataframe to be interpolated to the new index.
    new_index : pd.Index or np.array
        the new index.

    Returns
    -------
    df_out : pd.DataFrame
        a new dataframe interpolated to the new index.
    """
    df_out = pd.DataFrame(index=new_index)
    df_out.index.name = df.index.name

    for colname, col in df.items():
        if pd.api.types.is_numeric_dtype(col):
            df_out[colname + "_ip"] = np.interp(new_index, df.index, col)  # type: ignore

    return df_out


###############################################################################


def pd_Series_ip1d(
    src_df: pd.DataFrame,
    dst_df: pd.DataFrame,
    ivar_src_name: str,
    dvar_src_name: str,
    ivar_dst_name: str,
    dvar_dst_name: str,
    # see <https://docs.scipy.org/doc/scipy/reference/generated/scipy.interpolate.interp1d.html>:
    **kwargs,
) -> pd.DataFrame:
    """
    Interpolate dependent variable from source dataframe to destination df's independent variable.

    Modifies dst_df in-place!

    Uses numpy's interp function, wrapped by scipy.interpolate.interp1d

    Returns
    -------
    df_dst : pd.DataFrame
        modified input dst_df.
    """
    _df_dst = dst_df.copy()
    f = interp1d(
        src_df[ivar_src_name].values,
        src_df[dvar_src_name].values,
        **kwargs,
    )
    _df_dst[dvar_dst_name] = f(dst_df[ivar_dst_name])

    return _df_dst


###############################################################################


def pl_Series_interp1d(
    src_df: pl.DataFrame,
    dst_df: pl.DataFrame,
    ivar_src_name: str,
    dvar_src_name: str,
    ivar_dst_name: str,
    dvar_dst_name: str,
    # see <https://docs.scipy.org/doc/scipy/reference/generated/scipy.interpolate.interp1d.html>:
    **kwargs,
) -> pl.DataFrame:
    """
    Interpolate dependent variable from source dataframe to destination df's independent variable.

    Modifies dst_df in-place!

    Uses numpy's interp function, wrapped by scipy.interpolate.interp1d

    Returns
    -------
    df_dst : pl.DataFrame
        modified input dst_df.
    """
    f = interp1d(
        src_df[ivar_src_name].dt.timestamp(),
        src_df[dvar_src_name],
        **kwargs,
    )

    return dst_df.with_columns(
        pl.Series(f(dst_df[ivar_dst_name].dt.timestamp())).alias(dvar_dst_name)
    )


###############################################################################


def pl_Series_ip1d_lite(
    src_df: pl.DataFrame,
    dst_df: pl.DataFrame,
    ivar_src_name: str,
    dvar_src_name: str,
    ivar_dst_name: str,
    dvar_dst_name: str,
    # see <https://docs.scipy.org/doc/scipy/reference/generated/scipy.interpolate.interp1d.html>:
    **kwargs,
) -> pl.DataFrame:
    """
    Lite version of pl_Series_interp1d

    See https://stackoverflow.com/a/74819498/10197418
    """
    old = src_df[ivar_src_name].dt.timestamp().to_numpy()
    new = dst_df[ivar_dst_name].dt.timestamp().to_numpy()

    hi = np.searchsorted(old, new).clip(1, len(old) - 1)
    lo = hi - 1

    column = src_df[dvar_src_name].to_numpy()
    slope = (column[hi] - column[lo]) / (old[hi] - old[lo])

    out = pl.Series(slope * (new - old[lo]) + column[lo]).alias(dvar_dst_name)

    fill = kwargs.get("fill_value")
    if fill is not None and fill != "extrapolate":  # cannot use None as fill_value !
        m = (new > old.max()) | (new < old.min())
        out[m] = fill

    return dst_df.with_columns(out)



# SPDX-FileCopyrightText: 2025 Florian Obersteiner / KIT
# SPDX-FileContributor: Florian Obersteiner <f.obersteiner@kit.edu>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Corrections for time in timeseries data."""

import functools
from collections.abc import Callable
from copy import deepcopy
from datetime import timedelta
from typing import NamedTuple

import numpy as np
import polars as pl
import scipy as sc  # type:ignore
from matplotlib import pyplot as plt

###############################################################################


class CorrTime(NamedTuple):
    fitparms: np.ndarray
    t_corr: np.ndarray


def correct_time(t: np.ndarray, t_ref: np.ndarray, fitorder: int) -> CorrTime:
    """
    Fit a polynomial to the delta between a time vector and a reference time vector.

    time vector is corrected by subtracting the evaluated polynomial at each
        point of the time vector.
    inputs:
        t - time vector, 1D np array, numeric type
        t_ref - reference time vector, of same shape as t
        fitorder - order of the polynomial fit, integer

    Returns
    -------
        dict, holding
            'fitparms': parameters of the fit, ndarray
            't_corr': corrected input time vector t
    """
    try:
        parms = np.polyfit(t, t - t_ref, fitorder)
    except np.linalg.LinAlgError:  # sometimes happens at first try...
        parms = np.polyfit(t, t - t_ref, fitorder)
    t_corr = t - np.polyval(parms, t)

    return CorrTime(parms, t_corr)


###############################################################################


class FilteredDt(NamedTuple):
    n: int
    df: pl.DataFrame


def filter_dt_forward(df: pl.DataFrame, datetime_key: str = "datetime") -> FilteredDt:
    """
    Given a time series as polars.DataFrame, ensure that the index is increasing strictly.

    Filters forwards, i.e. if one element is less than the previous, then this
    element is removed (not the previous).

    Parameters
    ----------
    df : polars DataFrame
        DataFrame with datetime index.
    datetime_key : str, optional
        Date/time column name. The default is "datetime".

    Returns
    -------
    (n, df) : (int, polars DataFrame)
        The number of removed columns and the filtered dataframe.
    """
    fill_value = timedelta(microseconds=1)
    ref_value = timedelta(0)
    n_removed: int = 0

    m = df[datetime_key].diff().fill_null(fill_value) > ref_value
    while not m.all():
        if df.height == 1:
            return FilteredDt(n_removed, df)
        df = df.filter(m)
        n_removed += int((~m).sum())
        m = df[datetime_key].diff().fill_null(fill_value) > ref_value

    return FilteredDt(int(n_removed), df)


# -----------------------------------------------------------------------------


def filter_dt_backward(df: pl.DataFrame, datetime_key: str = "datetime") -> FilteredDt:
    """
    As filter_dt_forward, but backwards filtering.

    If one element is less than the previous, then the previous element is removed.
    """
    fill_value = timedelta(microseconds=1)
    ref_value = timedelta(0)
    n_removed: int = 0

    m = (df[datetime_key].diff().fill_null(fill_value) > ref_value).shift(-1).fill_null(True)
    while not m.all():
        if df.height == 1:
            return FilteredDt(n_removed, df)
        df = df.filter(m)
        n_removed += int((~m).sum())
        m = (df[datetime_key].diff().fill_null(fill_value) > ref_value).shift(-1).fill_null(True)

    return FilteredDt(n_removed, df)


###############################################################################


def xcorr_timelag(
    x1: np.ndarray,
    y1: np.ndarray,
    x2: np.ndarray,
    y2: np.ndarray,
    xrange: tuple[float, float] | None = None,
    upscale: int = 100,
    rmv_NaN: bool = True,
    pad_to_zero: bool = True,
    normalize_y: bool = True,
    # detrend_y=True, # possible future feature
    show_plots: bool = True,
    ynames: tuple[str, str] = ("f", "g"),
    corrmode: str = "positive",
    boundaries: tuple[float, float] | None = None,
    xcorr_func: Callable = sc.signal.correlate,
) -> float:
    """
    Analyze time lag between two time series f and g by cross-correlation.

    https://en.wikipedia.org/wiki/Cross-correlation#Time_delay_analysis

    a positive lag of g vs. f (reference) means that g lags behind in time.
    vice versa, if the lag is negative, g's signal preceedes that of f in time.

    Parameters
    ----------
    x1, y1 : 1d arrays
        independend and dependent variable of reference data.
    x2, y2 : 1d arrays
        independend and dependent variable of data to check for time lag.
    xrange : tuple, optional.
        cut data to fall within x-range "xrange". The default is (x1.min(), x2.max())
    upscale : numeric, scalar value
        upscale the data frequency by factor of upscale. The default is 100.
    rmv_NaN : boolean, optional
        clean NaNs from data. The default is True.
    pad_to_zero : boolean, optional
        drag the minimum of y to zero. The default is True.
    normalize_y : boolean, optional
        normalize y data to 0-1. The default is True.
    show_plots : boolean, optional
        show result plot. The default is True.
    ynames : tuple of str, optional
        2-element tuple of names to use on the plot. The default is ("f", "g").
    corrmode : str, optional.
        type of correlation to expect between y1 and y2. values: 'auto',
        'positive', 'negative'. The default is "positive".
    boundaries: 2-element tuple. lower and upper boundary.
        expect timelag to fall within these boundaries. The default is None.
    xcorr_func : func, optional
        function to calculate cross-correlation. The default is scipy.signal.correlate.

    Returns
    -------
    delay : float
        time delay of f vs. g in the unit of the input's independent variable.
    """
    # copy x and y so that nothing gets messed up
    x1, y1 = deepcopy(x1.astype(float)), deepcopy(y1.astype(float))
    x2, y2 = deepcopy(x2.astype(float)), deepcopy(y2.astype(float))

    # cut to selected xrange:
    if xrange is None:
        xrange = (x1.min(), x2.max())

    m1 = (x1 >= xrange[0]) & (x1 < xrange[1])
    x1, y1 = x1[m1], y1[m1]
    m2 = (x2 >= xrange[0]) & (x2 < xrange[1])
    x2, y2 = x2[m2], y2[m2]

    if rmv_NaN:
        m1 = np.isfinite(y1)
        y1, x1 = y1[m1], x1[m1]
        m2 = np.isfinite(y2)
        y2, x2 = y2[m2], x2[m2]

    if pad_to_zero:
        y1 -= np.nanmedian(y1)
        y2 -= np.nanmedian(y2)

    # this will fail if either one of the arrays holds no elements:
    if normalize_y:
        y1 /= np.nanmax(y1)
        y2 /= np.nanmax(y2)

    # normalize x:
    start, end = np.floor(x1[0]), np.ceil(x1[-1])
    n = (end - start) * upscale
    xnorm = np.linspace(start, end, num=int(n), endpoint=False)

    # interpolate y1 and y2 to xnorm:
    f_ip = sc.interpolate.interp1d(
        x1, y1, kind="linear", bounds_error=False, fill_value="extrapolate"
    )
    f = f_ip(xnorm)
    f_ip = sc.interpolate.interp1d(
        x2, y2, kind="linear", bounds_error=False, fill_value="extrapolate"
    )
    g = f_ip(xnorm)

    # cross-correlate f vs. g (i.e. y1 vs. y2):
    corr = xcorr_func(f, g)

    # need to know used correl function to make delay array...
    # check first if functools.partial was used.
    usedfunc = xcorr_func.func if xcorr_func.__class__ == functools.partial else xcorr_func  # type: ignore
    # if usedfunc.__code__ == np.correlate.__code__:
    #     delay_arr = np.linspace(-0.5 * n / upscale, 0.5 * n / upscale, int(n))[::-1]
    # elif usedfunc.__code__ == sc.signal.correlate.__code__:
    #     delay_arr = np.arange(1 - xnorm.size, xnorm.size) * (end - start) / xnorm.size * -1
    if usedfunc == np.correlate:
        delay_arr = np.linspace(-0.5 * n / upscale, 0.5 * n / upscale, int(n))[::-1]
    elif usedfunc == sc.signal.correlate:
        delay_arr = np.arange(1 - xnorm.size, xnorm.size) * (end - start) / xnorm.size * -1
    else:
        raise ValueError(f"unknown correl func: {usedfunc!r}")

    if boundaries:
        m = (delay_arr >= boundaries[0]) & (delay_arr < boundaries[1])
        delay_arr = delay_arr[m]
        corr = corr[m]

    # check if correlation is positive or negative to determine lag time
    assert corrmode in ("auto", "negative", "positive")
    if corrmode == "negative":
        select = np.argmin
    elif corrmode == "positive":
        select = np.argmax
    else:
        select = (np.argmin, np.argmax)[int(np.ceil(np.corrcoef(f, g)[0, 1]))]

    delay = delay_arr[select(corr)]

    if show_plots:
        _, ax = plt.subplots(2, 1, figsize=(14, 10))
        plt.subplots_adjust(top=0.94, bottom=0.06, left=0.06, right=0.94)
        p0 = ax[0].plot(x1, y1, "r", label=f"{ynames[0]}")
        ax[0].set_xlabel("x", weight="bold")
        ax[0].set_ylabel(f"{ynames[0]} normalized")
        ax[0].set_title("input")
        ax1 = ax[0].twinx()
        p1 = ax1.plot(x2, y2, "b", label=f"{ynames[1]}")
        ax1.set_ylabel(f"{ynames[1]} normalized")
        p2 = ax[0].plot(xnorm, f, "firebrick", label=f"{ynames[0]} resampled")  # type: ignore
        p3 = ax1.plot(xnorm, g, "deepskyblue", label=f"{ynames[1]} resampled")  # type: ignore
        plots = p0 + p1 + p2 + p3  # type: ignore
        lbls = [p.get_label() for p in plots]
        ax[0].legend(plots, lbls, loc=0, framealpha=1, facecolor="white")  # type: ignore
        ax[1].plot(delay_arr, corr, "k", label="xcorr")  # type: ignore
        ax[1].axvline(x=delay, color="r", linewidth=2)  # type: ignore
        ax[1].set_xlabel("lag")  # type: ignore
        ax[1].set_ylabel("correlation coefficient")  # type: ignore
        ax[1].set_title(f"{ynames[1]} vs. {ynames[0]} lag: {delay:+.3f}")  # type: ignore
        plt.show()

    return delay


###############################################################################


if __name__ == "__main__":
    # illustration of time_correction():
    order = 1
    t = np.array([1, 2, 3, 4, 5, 6], dtype=float)
    ref = np.array([1, 3, 5, 7, 9, 11], dtype=float)
    parms = np.polyfit(t, t - ref, order)
    t_corr = t - np.polyval(parms, t)
    plt.plot(t, "r", label="t")
    plt.plot(ref, "b", label="ref", marker="x")
    plt.plot(t_corr, "--k", label="corrected")
    plt.plot(t - ref, "g", label="delta (t-ref)")
    plt.legend()



# SPDX-FileCopyrightText: 2025 Florian Obersteiner / KIT
# SPDX-FileContributor: Florian Obersteiner <f.obersteiner@kit.edu>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Filtering and Masking."""

from typing import NamedTuple

import numpy as np
import numpy.typing as npt
from numba import njit  # type:ignore
from scipy.interpolate import interp1d  # type:ignore
from sklearn.neighbors import LocalOutlierFactor  # type:ignore

###############################################################################


def mask_repeated(a: np.ndarray, N: int, atol: float = 1e-6) -> np.ndarray:
    """
    Mask elements that repeat more than N times.

    on SO:
        https://stackoverflow.com/a/58482894/10197418

    Parameters
    ----------
    a : 1d array
    N : int
        mask element if it repeats more than n times
    atol : float, optional
        absolute tolerance to check for equality. The default is 1e-6.

    Returns
    -------
    boolean mask

    """
    mask = np.ones(a.shape[0], np.bool_)
    mask[N:] = ~np.isclose(a[N:], a[:-N], atol=atol, equal_nan=True)

    return mask


###############################################################################


@njit
def mask_repeated_nb(arr: np.ndarray, n: int, atol: float = 1e-6) -> np.ndarray:
    """
    Mask elements that repeat more than N times, njit'ed version.

    Parameters
    ----------
    arr : 1d array
    n : int
        mask element if it repeats more than n times
    atol : float, optional
        absolute tolerance to check for equality. The default is 1e-6.

    Returns
    -------
    boolean mask
    """
    mask = np.ones(arr.shape, np.bool_)
    current = arr[0]
    count = 0
    for idx, item in enumerate(arr):
        if abs(item - current) < atol:
            count += 1
        else:
            current = item
            count = 1
        mask[idx] = count <= n

    return mask


###############################################################################


@njit
def mask_jumps(
    arr: np.ndarray,
    threshold: float | int,
    look_ahead: int,
    use_abs_delta: bool = False,
) -> np.ndarray:
    """
    Check elements of array "arr" if difference between subsequent elements exceeds threshold.

    Parameters
    ----------
    arr : np.ndarray
        Numpy 1darray to analyze.
    threshold : [float, int]
        Maximum allowed difference between subsequent elements.
    look_ahead : int
        How many elements to look ahead if a difference exceeds threshold.
    use_abs_delta : bool, optional
        Consider the absolute difference. The default is False.

    Returns
    -------
    mask : np.ndarray
        Boolean mask.
    """
    n_el = arr.shape[0]
    mask = np.ones(arr.shape).astype(np.bool_)
    i = 0
    while i < n_el - 1:
        cur, nxt = arr[i], arr[i + 1]
        delta_0 = np.absolute(nxt - cur) if use_abs_delta else nxt - cur
        if delta_0 > threshold:
            for value in arr[i + 1 : i + look_ahead + 1]:
                delta_1 = np.absolute(value - cur) if use_abs_delta else value - cur
                if delta_1 > threshold:
                    mask[i + 1] = False
                    i += 1
                else:
                    break
        i += 1

    return mask


###############################################################################


class JumpfilterResult(NamedTuple):
    filtered: np.ndarray
    mask: np.ndarray


def filter_jumps(
    arr: npt.NDArray[np.float64],
    threshold: float,
    look_ahead: int,
    use_abs_delta: bool = False,
    vmiss_idx: npt.NDArray[np.int64] | None = None,
    remove_repeated: bool = False,
    interpol_jumps: bool = False,
    interpol_kind: str = "linear",
) -> JumpfilterResult:
    """
    Wrap mask_jumps(). Only works for floating point number arrays.

    (!) interpolation assumes equidistant spacing of the independent variable on
      which arr depends.
    """
    if arr.ndim > 1:
        raise ValueError("input array must be numpy 1d array.")
    if look_ahead >= arr.shape[0] or look_ahead < 1:
        raise ValueError(f"parameter look_ahead must be >=1 and <{arr.shape[0]}.")

    result = arr.copy()  # do not touch the input...
    if vmiss_idx is not None:
        result[vmiss_idx] = np.nan

    if remove_repeated:
        result[~mask_repeated(result, 2)] = np.nan

    mask = mask_jumps(result, threshold, look_ahead, use_abs_delta=use_abs_delta)
    result[~mask] = np.nan
    if interpol_jumps:
        f_ip = interp1d(
            np.arange(0, result.shape[0])[mask],
            result[mask],
            kind=interpol_kind,
            fill_value="extrapolate",  # type: ignore
        )
        result = f_ip(np.arange(0, result.shape[0]))

    return JumpfilterResult(result, mask)


###############################################################################


class JumpfilterResult2(NamedTuple):
    filtered: np.ndarray
    deleted_idx: np.ndarray
    remaining_idx: np.ndarray


def filter_jumps_np(
    v: npt.NDArray[np.float64],
    max_delta: float,
    no_val: float = np.nan,
    use_abs_delta: bool = True,
    reset_buffer_after: int = 3,
    remove_doubles: bool = False,
    interpol_jumps: bool = False,
    interpol_kind: str = "linear",
) -> JumpfilterResult2:
    """
    Mask jumps using numpy functions.

    If v is dependent on another variable x (e.g. time) and if that x
    is not equidistant, do NOT use interpolation.

    Parameters
    ----------
    v : np 1d array
        data to filter.
    max_delta : float
        defines "jump".
    no_val : float, optional
        missing value placeholder. The default is np.nan.
    use_abs_delta : boolean, optional
        use the absolute delta to identify jumps. The default is True.
    reset_buffer_after : int, optional
        how many elements to wait until reset. The default is 3.
    remove_doubles : boolean, optional
        remove elements that are repeated once. The default is False.
    interpol_jumps : boolean, optional
        decide to interpolate filtered values. The default is False.
    interpol_kind : string, optional
        how to interpolate, see scipy.interpolate.interp1d.
        The default is 'linear'.

    Returns
    -------
    NamedTuple, .filtered: filtered data
                .deleted_idx: indices of deleted elements
                .remaining_idx: indices of remaining elements

    """
    ix_del = np.full(v.shape[0], -1, dtype=int)  # deletion index
    ix_rem = np.full(v.shape[0], -1, dtype=int)  # remaining index

    buffer = [False, 0]

    for ix, v_ix in enumerate(v):
        if any([~np.isfinite(v_ix), v_ix == no_val, np.isnan(v_ix)]):
            ix_rem[ix] = ix
            continue  # skip line if value is np.nan

        if not buffer[0]:
            buffer[0] = v_ix  # type:ignore
            ix_rem[ix] = ix
            continue  # fill buffer if not done so yet

        delta = abs(v_ix - buffer[0]) if use_abs_delta else v_ix - buffer[0]

        if delta > max_delta:  # jump found!
            v[ix] = no_val
            ix_del[ix] = ix
            buffer[1] += 1
            if reset_buffer_after and buffer[1] == reset_buffer_after:
                buffer = [v_ix, 0]  # type:ignore
        else:  # no jump,...
            buffer[0] = v_ix  # type:ignore
            if remove_doubles:  # check for double values...
                if np.isclose(delta, 0.0):  # double found!
                    v[ix] = no_val
                    ix_del[ix] = ix
                else:  # no double
                    ix_rem[ix] = ix
            else:
                ix_rem[ix] = ix

    ix_del = ix_del[np.where(ix_del != -1)]  # type: ignore

    ix_rem = ix_rem[np.where(ix_rem != -1)]  # type: ignore

    if interpol_jumps:
        tmp_x = (np.arange(0, v.shape[0]))[ix_rem]
        tmp_y = v[ix_rem]
        f_ip = interp1d(tmp_x, tmp_y, kind=interpol_kind, fill_value="extrapolate")  # type: ignore
        filtered = f_ip(np.arange(0, v.shape[0]))
    else:
        filtered = v[ix_rem]

    return JumpfilterResult2(filtered, ix_del, ix_rem)


###############################################################################


def extend_mask(m: np.ndarray, n: int) -> np.ndarray:
    """
    extend_mask changes elements from False to True around existing 'True' elements.

    Parameters
    ----------
    m : np.ndarray
        boolean mask.
    n : int
        number of 'True' elements to insert.
        'True' is inserted right and left of existing 'True' elements in alternating fashion.

    Returns
    -------
    np.ndarray
        updated mask.

    """
    return np.convolve(m, np.ones((n + 1,)), mode="same").astype(np.bool_)


###############################################################################


def simple_1d_lof(var: np.ndarray, k: int, thresh: float, mode="both") -> np.ndarray:
    """
    simple_1d_lof runs an outlier detection based on the local outlier factor algorithm,
      https://en.wikipedia.org/wiki/Local_outlier_factor
    The application to a time series is derived from the 'UniLOF' flagging scheme from SaQC,
      https://rdm-software.pages.ufz.de/saqc/

    Parameters
    ----------
    var : np.array
        dependent variable of the time series, to be analyzed with lof.
    k : int
        number of neighbors to use for the distance calculation.
    thresh : float
        the threshold value to determine if a value is tagged as an outlier.
        a typical value is 1.5
    mode : str, optional
        mark only 'positive', 'negative' outliers or 'both' (default)

    Returns
    -------
    np.ndarray
        boolean mask that will be True where an outlier is detected.

    """
    assert len(var.shape) == 1, "can only work with 1D data"
    assert np.isfinite(var).all(), "can only use all-finite values in 'var'"

    v_diff = np.diff(var, prepend=var[0])

    density = np.median(np.abs(v_diff))
    if density == 0:
        m = v_diff != 0
        density = 1 if not m.any() else np.median(np.abs(v_diff[m]))

    d_var = np.arange(var.shape[0]) * density

    # prepend / append k elements to var to detect outliers on the edges
    extd_var = np.pad(var, k, mode="reflect")
    extd_dens = np.pad(d_var, k, mode="reflect", reflect_type="odd")

    clf = LocalOutlierFactor(n_neighbors=k)
    _ = clf.fit_predict(np.array([extd_var, extd_dens]).T)
    resids = clf.negative_outlier_factor_[k:-k]  # truncate what was pre-/appended

    m = np.absolute(resids) > thresh

    if mode == "positive":
        return m & (v_diff > 0)
    if mode == "negative":
        return m & (v_diff < 0)
    return m


###############################################################################



# SPDX-FileCopyrightText: 2025 Florian Obersteiner / KIT
# SPDX-FileContributor: Florian Obersteiner <f.obersteiner@kit.edu>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import sys
import tomllib

PYPROJECT_PATH = "pyproject.toml"  # assume script runs in root


def get_version() -> str:
    """
    Reads the version from the pyproject.toml file.
    Supports both the standard [project] section and [tool.poetry].
    """
    try:
        with open(PYPROJECT_PATH, "rb") as fd:
            data = tomllib.load(fd)

            # Try the standard [project] section first (PEP 621)
            if "project" in data and "version" in data["project"]:
                return data["project"]["version"]

            # Then try the [tool.poetry] section (if using Poetry)
            elif (
                "tool" in data and "poetry" in data["tool"] and "version" in data["tool"]["poetry"]
            ):
                return data["tool"]["poetry"]["version"]

            else:
                print(
                    f"Error: Could not find version field in [project] or [tool.poetry] in {PYPROJECT_PATH}",
                    file=sys.stderr,
                )
                sys.exit(1)

    except FileNotFoundError:
        print(f"Error: {PYPROJECT_PATH} not found.", file=sys.stderr)
        sys.exit(1)
    except tomllib.TOMLDecodeError as e:
        print(f"Error parsing {PYPROJECT_PATH}: {e}", file=sys.stderr)
        sys.exit(1)
    except Exception as e:
        print(f"An unexpected error occurred: {e}", file=sys.stderr)
        sys.exit(1)


if __name__ == "__main__":
    version = get_version()
    if version:
        print(version)



# SPDX-FileCopyrightText: 2025 Florian Obersteiner / KIT
# SPDX-FileContributor: Florian Obersteiner <f.obersteiner@kit.edu>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import unittest
from pathlib import Path

import numpy as np

from pyfuppes import misc


class TestTimecorr(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # to run before all tests
        pass

    @classmethod
    def tearDownClass(cls):
        # to run after all tests
        pass

    def setUp(self):
        # to run before each test
        pass

    def tearDown(self):
        # to run after each test
        pass

    def test_listelement(self):
        a = [1, 2, 3]
        b = misc.list_change_elem_index(a, 3, 0)
        self.assertListEqual(b, [3, 1, 2])
        self.assertListEqual(a, [1, 2, 3])
        with self.assertRaises(IndexError):
            _ = misc.list_change_elem_index(a, 3, 5)

    def test_pathcleaner(self):
        p = "$HOME"
        clean_p = misc.clean_path(p)
        self.assertEqual(clean_p.as_posix(), Path().home().as_posix())
        p = "~"
        clean_p = misc.clean_path(p)
        self.assertEqual(clean_p.as_posix(), Path().home().as_posix())
        p = "$HOME/a_directory"
        clean_p = misc.clean_path(p)
        self.assertEqual(clean_p.as_posix(), (Path().home() / "a_directory").as_posix())

    def test_findfirst(self):
        arr = np.arange(10)
        elem = misc.find_fist_elem(arr, 4, lambda x, y: x > y)
        self.assertEqual(5, elem)
        lst = list(range(10))
        elem = misc.find_fist_elem(lst, 4, lambda x, y: x > y)
        self.assertEqual(5, elem)


if __name__ == "__main__":
    unittest.main()



# SPDX-FileCopyrightText: 2025 Florian Obersteiner / KIT
# SPDX-FileContributor: Florian Obersteiner <f.obersteiner@kit.edu>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import unittest

import numpy as np

from pyfuppes import monotonicity


class TestMonotonicity(unittest.TestCase):
    def test_strictly_increasing(self):
        self.assertTrue(monotonicity.strictly_increasing([1, 2, 3]))
        self.assertFalse(monotonicity.strictly_increasing([1, 2, 2]))
        self.assertFalse(monotonicity.strictly_increasing([1, 1, 3]))

    def test_strictly_decreasing(self):
        self.assertTrue(monotonicity.strictly_decreasing([3, 2, 1]))
        self.assertFalse(monotonicity.strictly_decreasing([3, 2, 2]))
        self.assertFalse(monotonicity.strictly_decreasing([3, 3, 1]))

    def test_non_increasing(self):
        self.assertTrue(monotonicity.non_increasing([1, 1, 1]))
        self.assertTrue(monotonicity.non_increasing([1, 1, -2]))
        self.assertFalse(monotonicity.non_increasing([1, 1, 3]))

    def test_non_decreasing(self):
        self.assertTrue(monotonicity.non_decreasing([1, 2, 3]))
        self.assertTrue(monotonicity.non_decreasing([1, 2, 2]))
        self.assertTrue(monotonicity.non_decreasing([2, 2, 2]))
        self.assertFalse(monotonicity.non_decreasing([1, 1, -3]))

    def test_strictly_inc_np(self):
        self.assertTrue(monotonicity.strict_inc_np(np.array([1, 2, 3])))
        self.assertFalse(monotonicity.strict_inc_np(np.array([1, 2, 2])))
        self.assertFalse(monotonicity.strict_inc_np(np.array([1, 1, 3])))

    def test_strictly_dec_np(self):
        self.assertTrue(monotonicity.strict_dec_np(np.array([3, 2, 1])))
        self.assertFalse(monotonicity.strict_dec_np(np.array([3, 2, 2])))
        self.assertFalse(monotonicity.strict_dec_np(np.array([3, 3, 1])))

    def test_non_inc_np(self):
        self.assertTrue(monotonicity.non_inc_np(np.array([2, 2, 2])))
        self.assertFalse(monotonicity.non_inc_np(np.array([1, 2, 2])))
        self.assertFalse(monotonicity.non_inc_np(np.array([1, 1, 3])))

    def test_non_dec_np(self):
        self.assertTrue(monotonicity.non_dec_np(np.array([1, 2, 3])))
        self.assertTrue(monotonicity.non_dec_np(np.array([2, 2, 2])))
        self.assertFalse(monotonicity.non_dec_np(np.array([1, -2, 2])))
        self.assertFalse(monotonicity.non_dec_np(np.array([1, 1, -3])))


if __name__ == "__main__":
    unittest.main()



# SPDX-FileCopyrightText: 2025 Florian Obersteiner / KIT
# SPDX-FileContributor: Florian Obersteiner <f.obersteiner@kit.edu>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Convert between types representing date/time."""

import warnings
from datetime import UTC, datetime, timedelta, timezone
from operator import attrgetter
from typing import Any

import numpy as np
import numpy.typing as npt
import xarray as xr

### HELPERS ###################################################################


def _to_list(parm: Any | list[Any] | np.ndarray) -> tuple[list | np.ndarray, bool]:
    """
    Convert input "parm" to a Python list object.

    If "parm" is a scalar, return value "is_scalar" is True, otherwise False.
    """
    if isinstance(parm, str):  # check this first: don't call list() on a string
        parm = [parm]
    if isinstance(parm, np.ndarray):
        parm = parm.tolist()

    # 'anything else but not list' case:
    if not isinstance(parm, list):
        try:
            parm = list(parm)  # call list() first in case parm is np.ndarray
        except TypeError:  # will e.g. raised if parm is a float
            parm = [parm]
    # else: we have a list already

    return parm, len(parm) == 1


### MAIN FUNCTIONS ############################################################


def xrtime_to_mdns(xrda: xr.DataArray, dim_name="Time") -> npt.NDArray[np.float64]:
    """
    Convert the time vector of an xarray.DataArray to an array representing seconds after midnight.

    Parameters
    ----------
    xrda : xr.DataArray
        xarray.DataArray to extract time from.
    dim_name : str, optional
        Attribute name of the time dimension. The default is "Time".

    Returns
    -------
    np.array
        time in seconds after midnight (dtype float).

    """
    f = attrgetter(dim_name)
    t = f(xrda)

    return (t - t[0].dt.floor("d")).values.astype(int) / 1_000_000_000


###############################################################################


def dtstr_2_mdns(
    timestring: str | list,
    tsfmt: str = "%Y-%m-%d %H:%M:%S.%f",
    ymd: tuple[int, ...] | None = None,
) -> float | list[float]:
    """
    Convert datetime string to seconds since midnight (float).

    Since a relative difference is calculated, the function is 'timezone-safe'.
    Variable UTC offsets are not allowed.

    Parameters
    ----------
    timestring : str, list of str or np.ndarray with dtype str/obj.
        timestamp given as string.
    tsfmt : str, optional
        timestring format. Pass "iso" to denote ISO8601 format. The default is "%Y-%m-%d %H:%M:%S.%f".
    ymd : tuple, optional
        starting date as tuple of integers; (year, month, day).
        The default is None.

    Returns
    -------
    float; scalar or float; list
        seconds since midnight for the given timestring(s).
    """
    _timestrings, ret_scalar = _to_list(timestring)

    if tsfmt == "iso":
        dts = [datetime.fromisoformat(s) for s in _timestrings]
    else:
        dts = [datetime.strptime(s, tsfmt) for s in _timestrings]

    return dtobj_2_mdns(dts[0] if ret_scalar else dts, ref_date=ymd, ref_is_first=True)


###############################################################################


def dtobj_2_mdns(
    dt_obj: datetime | list[datetime],
    ref_date: tuple[int, ...] | None = None,
    ref_is_first: bool = False,
) -> float | list[float]:
    """
    Convert a Python datetime object (or list/array of ...) to seconds after midnight.

    Only a single timezone or no timezone (naive datetime) is allowed.

    Parameters
    ----------
    dt_obj : datetime object or list/array of datetime objects
        the datetime to be converted to seconds after midnight.
    ref_date : tuple of int, optional
        custom start date given as (year, month, day). The default is False.
    ref_is_first : bool, optional
        first entry of dt_obj list/array defines start date.
        The default is False.

    Returns
    -------
    float; scalar or list of float
        seconds after midnight for the given datetime object(s).
    """
    _dt_objs, ret_scalar = _to_list(dt_obj)

    tzs = [d.tzinfo for d in _dt_objs]
    assert len(set(tzs)) == 1, "all time zones (tzinfo) must be equal."

    t0 = _dt_objs[0]
    if ref_date:
        t0 = datetime(*ref_date, tzinfo=_dt_objs[0].tzinfo)  # type: ignore
    t0 = t0.replace(hour=0, minute=0, second=0, microsecond=0)

    if ref_is_first or ref_date:
        result = [(x - t0).total_seconds() for x in _dt_objs]
    else:
        result = [
            (x - x.replace(hour=0, minute=0, second=0, microsecond=0)).total_seconds()
            for x in _dt_objs
        ]

    return result[0] if ret_scalar else result


###############################################################################


def unixtime_2_mdns(
    timestamp: float | list[float], ymd: tuple[int, ...] | None = None
) -> float | list[float]:
    """
    Convert UNIX time (or list/array of ...) to seconds after midnight.

    Parameters
    ----------
    posixts : float, list of float or np.ndarray with dtype float.
        the POSIX timestamp to be converted to seconds after midnight.
    ymd : tuple of int, optional
        define starting date as tuple of integers (year, month, day) UTC.
        The default is None, which means the reference date is that of the
        first element in posixts.

    Returns
    -------
    float; scalar or list of float
        seconds after midnight for the given POSIX timestamp(s).
    """
    _timestamps, ret_scalar = _to_list(timestamp)

    # to floor a Unix time to the date, use  t - t % 86400
    # here, we need to account for the fact that the reference date might be different.

    if ymd:  # (yyyy, m, d) given, take that as starting point t0:
        t0 = datetime(year=ymd[0], month=ymd[1], day=ymd[2], tzinfo=UTC).timestamp()
    else:  # take date of first entry as starting point
        t0_dt = datetime.fromtimestamp(_timestamps[0], tz=UTC)
        t0 = t0_dt.replace(hour=0, minute=0, second=0, microsecond=0).timestamp()

    ts = [t - t0 for t in _timestamps]

    return ts[0] if ret_scalar else ts


###############################################################################


def mdns_2_dtobj(
    mdns: float | list[float],
    ref_date: tuple[int] | datetime,
    assume_UTC: bool = True,
    posix: bool = False,
    str_fmt: str = "",
) -> datetime | float | str | list[datetime] | list[float] | list[str]:
    """
    Convert seconds after midnight (or list/array of ...) to datetime object.

    Parameters
    ----------
    mdns : float, list of float or np.ndarray with dtype float.
        the seconds after midnight to be converted to datetime object(s).
    ref_date : tuple of int (year, month, day) or datetime object
        date that mdns refers to.
    assume_UTC : boolean.
        if ref_date is supplied as a y/m/d tuple, add tzinfo UTC.
    posix : bool, optional
        return POSIX timestamp(s). The default is False.
    str_fmt : str, optional
        Format for datetime.strftime, e.g. "%Y-%m-%d %H:%M:%S.%f"
        If provided, output is delivered as formatted string. POSIX must
            be False in that case, or STR_FMT is overridden (evaluated last).
        The default is False.

    Returns
    -------
    datetime object or float (POSIX timestamp)
        ...for the given seconds after midnight.
    """
    _mdnsecs, ret_scalar = _to_list(mdns)
    # ensure type float:
    if not isinstance(_mdnsecs[0], float | np.floating):
        _mdnsecs = list(map(float, _mdnsecs))

    # check if ref_date is supplied as a y/m/d tuple. convert to datetime.
    reset_tz = False
    if isinstance(ref_date, tuple | list):
        ref_dt, reset_tz = datetime(*ref_date), True  # type:ignore
    else:
        ref_dt = ref_date

    if assume_UTC:  # add timezone UTC if assume_UTC is set to True
        ref_dt = ref_dt.replace(tzinfo=UTC)

    result: list[datetime] = [ref_dt + timedelta(seconds=t) for t in _mdnsecs]  # type:ignore

    if posix:
        if not ref_dt.tzinfo:
            print(
                "*mdns_2_dtobj warning*: creating POSIX timestamps from "
                "naive datetime objects might give unexpected results!\n"
                "\t-> consider passing a tz-aware ref_date instead."
            )
        result: list[float] = [dtobj.timestamp() for dtobj in result]  # type: ignore
    elif str_fmt:
        offset = -3 if str_fmt.endswith("%f") else None
        result: list[str] = [dtobj.strftime(str_fmt)[:offset] for dtobj in result]  # type: ignore
    else:
        if reset_tz:
            result: list[datetime] = [t.replace(tzinfo=None) for t in result]  # type: ignore

    return result[0] if ret_scalar else result


###############################################################################


def daysSince_2_dtobj(day0: datetime, days_since: int | float) -> datetime | list[datetime]:
    """
    Convert a date and a floating point number "days_since" to a datetime object.

    day0: datetime object, from when to count.

    Parameters
    ----------
    day0 : datetime object (naive or tz-aware)
        from when to count.
    daysSince : int or float
        number of days.

    Returns
    -------
    datetime object
    """
    if isinstance(days_since, list | np.ndarray):
        return [(day0 + timedelta(days=ds)) for ds in days_since]
    return day0 + timedelta(days=days_since)


###############################################################################


def dtstr_2_unixtime(
    timestring: str, tsfmt: str = "%Y-%m-%d %H:%M:%S.%f", tz: timezone = UTC
) -> float:
    """
    Convert timestring without time zone information to Unix time.

    Parameters
    ----------
    timestring : string
        representing date (and time).
    tsfmt : str, optional
        strptime format. The default is "%Y-%m-%d %H:%M:%S.%f".
        Set to 'iso' to use Python's datetime.fromisoformat() method.
    tz : timezone, optional
        The default is timezone.utc for UTC.
        Set to None to ignore/use tzinfo as parsed.
        Note: if tzinfo is None after parsing, and tz argument is None,
            Python will assume local time by default!

    Returns
    -------
    POSIX timestamp / UNIX time
        UTC seconds since the Unix epoch 1970-01-01.
    """
    if tsfmt == "iso":
        dtobj = datetime.fromisoformat(timestring)
    else:
        dtobj = datetime.strptime(timestring, tsfmt)

    # if parsed dtobj neither has tzinfo nor utcoffset defined...
    if dtobj.tzinfo is None and dtobj.utcoffset() is None:
        if tz:  # set the tzinfo if tz argument is provided
            dtobj = dtobj.replace(tzinfo=tz)
        else:  # issue a warning if tz argument is None
            warnings.warn(
                "Warning: ambiguous datetime - parsed datetime object won't have tzinfo set!",
                stacklevel=2,
            )
    # else the dtobj already has a tz, so convert if tz argument is given:
    else:
        if tz:
            dtobj = dtobj.astimezone(tz)

    return dtobj.timestamp()


###############################################################################



# SPDX-FileCopyrightText: 2025 Florian Obersteiner / KIT
# SPDX-FileContributor: Florian Obersteiner <f.obersteiner@kit.edu>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Geospatial helpers, such as Haversine distance or solar zenith angle."""

from datetime import datetime
from math import acos, asin, cos, degrees, floor, radians, sin, sqrt

import numpy as np
from geopy import distance  # type:ignore
from numba import njit  # type:ignore
from pysolar.solar import get_altitude  # type:ignore


# trigonometry with degrees:
def _cos2(x):
    return cos(radians(x))


def _sin2(x):
    return sin(radians(x))


def _acos2(x):
    return degrees(acos(x))


###############################################################################

EARTH_RADIUS: float = 6372.8  # approximate radius of earth in km


@njit
def haversine_dist(lat: np.ndarray, lon: np.ndarray) -> float:
    """
    Calculate the great-circle distance between two sets of coordinates
    using the Haversine formula.

    This function computes the distance between two sets of latitude and
    longitude coordinates, assuming a spherical Earth.  It's optimized
    for performance using NumPy and Numba.

    Args:
        lat: A NumPy array of latitude values (in degrees).
        lon: A NumPy array of longitude values (in degrees).

    Returns:
        A float representing the total distance in kilometers between the
        two sets of coordinates.

    Raises:
        TypeError: If either `lat` or `lon` is not a NumPy array.
        ValueError: If the lengths of `lat` and `lon` are not equal.
    """
    assert lat.shape[0] == lon.shape[0], "lat/lon must be of same length."
    dist: float = 0

    lat1, lon1 = lat[0], lon[0]
    for j in range(1, lat.shape[0]):
        lat0, lat1 = lat1, lat[j]
        lon0, lon1 = lon1, lon[j]

        dLat = radians(lat1 - lat0)
        dLon = radians(lon1 - lon0)
        lat0 = radians(lat0)
        lat1 = radians(lat1)

        a = sin(dLat / 2) ** 2 + cos(lat0) * cos(lat1) * sin(dLon / 2) ** 2
        c = 2 * asin(sqrt(a))

        dist += EARTH_RADIUS * c

    return dist


###############################################################################


def geodesic_dist(lat: np.ndarray, lon: np.ndarray, ellipsoid="WGS-84") -> np.ndarray:
    """
    Calculate geodesic distance in km along lat/lon coordinates. Wraps geopy.distance.geodesic.

    Returns
        array of step-wise distance for each pair of lat/lon coordinates
    """
    assert lat.shape[0] == lon.shape[0], "lat/lon must be of same length."

    result = np.zeros(lat.shape, dtype=float)
    for j in range(lat.shape[0] - 1):
        result[j + 1] = distance.geodesic(
            (lat[j], lon[j]), (lat[j + 1], lon[j + 1]), ellipsoid=ellipsoid
        ).km

    return result


###############################################################################


def sza_pysolar(
    UTC: datetime,
    latitude: float,
    longitude: float,
) -> float:
    """Compute solar zenith angle with get_altitude function from pysolar package."""
    return 90 - get_altitude(latitude, longitude, UTC)


# another option to calculate SZA would be to use astropy; for instance like
#
# from astropy.coordinates import get_sun, AltAz, EarthLocation
# from astropy.time import Time
#
# lon, lat = 42, 55
# sun_time = Time('2017-12-6 17:00') # UTC
# loc = EarthLocation.from_geodetic(lon, lat)
# altaz = AltAz(obstime=sun_time, location=loc)
# zen_ang = get_sun(sun_time).transform_to(altaz).zen
#
# see e.g. https://stackoverflow.com/a/47678091/10197418


def sza(
    UTC: datetime,
    latitude: float,
    longitude: float,
) -> float:
    """
    Calculate the solar zenith angle (in degree).

    UTC         (as datetime.datetime Object)
    longitude   (in degree)
    latitude    (in degree)

    Code adapted from https://www.python-forum.de/viewtopic.php?t=21117
    (2018-10-17 8:10 UTC)
    """

    day_of_year = UTC.timetuple().tm_yday
    leap_year_factor = (-0.375, 0.375, -0.125, 0.125)[UTC.year % 4]
    UTC_min = UTC.hour * 60.0 + UTC.minute + UTC.second / 60.0
    J = 360.0 / 365.0 * (day_of_year + leap_year_factor + UTC_min / 1440.0)

    # hour angle (using the time equation)
    average_localtime = UTC_min + 4 * longitude
    true_solar_time = (
        average_localtime
        + 0.0066
        + 7.3525 * _cos2(J + 85.9)
        + 9.9359 * _cos2(2 * J + 108.9)
        + 0.3387 * _cos2(3 * J + 105.2)
    )
    hour_angle = 15.0 * (720.0 - true_solar_time) / 60.0

    # sun declination, using DIN 5034-2
    declination = (
        0.3948
        - 23.2559 * _cos2(J + 9.1)
        - 0.3915 * _cos2(2 * J + 5.4)
        - 0.1764 * _cos2(3 * J + 26.0)
    )

    # solar zenith angle
    return _acos2(
        _sin2(latitude) * _sin2(declination)
        + _cos2(hour_angle) * _cos2(latitude) * _cos2(declination)
    )


###############################################################################


def get_EoT(date_ts: datetime) -> float:
    """
    Calculate equation of time, for calculation of local solar time.

    input: date_ts - datetime object

    returns: equation of time result - float
    """
    B = (360 / 365) * (date_ts.timetuple().tm_yday - 81)

    return 9.87 * sin(2 * B) - 7.53 * cos(B) - 1.5 * sin(B)


###############################################################################


def get_LSTdayFrac(
    longitude: float, tz_offset: float, EoT: float, days_delta: float, time_delta: float
) -> float:
    """
    Calculate local solar time as a fraction of a day.

    input:
        longitude: -180 to +180 degrees west to east, float
        tz_offset: time zone offset from UTC in hours, float
        EoT: equation of time for selected date, float
        days_delta: days since reference date, float
        time_delta: current time as days since reference date, float
    returns:
        local solar time as day fraction (float, 0-1)
    """
    LSTM = 15 * tz_offset  # Local Standard Time Meridian
    t_corr = (4 * (longitude - LSTM) + EoT) / 60 / 24  # [d]
    LST_frac = (time_delta + tz_offset / 24 - days_delta) + t_corr
    if LST_frac > 1:
        return LST_frac - floor(LST_frac)
    return LST_frac



# SPDX-FileCopyrightText: 2025 Florian Obersteiner / KIT
# SPDX-FileContributor: Florian Obersteiner <f.obersteiner@kit.edu>
#
# SPDX-License-Identifier: GPL-3.0-or-later

from importlib import metadata

from pyfuppes import (
    avgbinmap,
    dictstuff,
    filters,
    geo,
    interpolate,
    misc,
    monotonicity,
    na1001,
    numberstring,
    plottools,
    timeconversion,
    timecorr,
    txt2dict,
)

__version__ = metadata.version("pyfuppes")

__all__ = (
    "avgbinmap",
    "dictstuff",
    "filters",
    "geo",
    "interpolate",
    "misc",
    "monotonicity",
    "na1001",
    "numberstring",
    "plottools",
    "timeconversion",
    "timecorr",
    "txt2dict",
)



# SPDX-FileCopyrightText: 2025 Florian Obersteiner / KIT
# SPDX-FileContributor: Florian Obersteiner <f.obersteiner@kit.edu>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Dictionary tools."""

import tomllib
from pathlib import Path
from typing import Any, Self

# --------------------------------------------------------------------------------------------------


class DotDict(dict):
    """Wrapper class around dict for dot-notation access to dictionary keys."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize DotDict and recursively convert nested dictionaries."""
        super().__init__(*args, **kwargs)
        self._convert_nested()

    def _convert_nested(self, path: str = "") -> None:
        """Convert nested dictionaries to DotDict instances and validate keys."""
        for key, value in list(self.items()):
            # Validate keys
            if not isinstance(key, str):
                continue

            # Check for invalid attribute names
            if not key.isidentifier():
                print(
                    f"Warning: '{path}{key}' is not a valid Python identifier and cannot be accessed with dot notation"
                )

            # Check for reserved dictionary method names
            if key in dir(dict) and key not in ("__class__", "__dict__"):
                print(f"Warning: '{path}{key}' overrides a built-in dictionary method")

            # Recursively convert nested dictionaries
            if isinstance(value, dict) and not isinstance(value, DotDict):
                self[key] = DotDict(value)
                self[key]._convert_nested(f"{path}{key}.")
            elif isinstance(value, list):
                # Handle lists and recursively process any dictionaries within them
                self[key] = self._process_list_items(value, f"{path}{key}")

    def _process_list_items(self, items: list[Any], path: str) -> list[Any]:
        """Recursively process items in a list, converting any dicts to DotDict."""
        processed_items = []
        for i, item in enumerate(items):
            if isinstance(item, dict) and not isinstance(item, DotDict):
                dot_dict_item = DotDict(item)
                dot_dict_item._convert_nested(f"{path}[{i}].")
                processed_items.append(dot_dict_item)
            elif isinstance(item, list):
                # Handle nested lists recursively
                processed_items.append(
                    self._process_list_items(item, f"{path}[{i}]")  # type:ignore
                )  # type:ignore
            else:
                processed_items.append(item)
        return processed_items

    @classmethod
    def from_toml_data(cls, toml_data: dict[str, Any]) -> Self:
        """Create a DotDict instance from TOML data."""
        return cls(toml_data)

    @classmethod
    def from_toml_file(cls, file_path: str | Path) -> Self:
        """Load a TOML file into a DotDict instance."""
        with open(file_path, "rb") as f:
            config_data = tomllib.load(f)
        return cls(config_data)

    def __getattr__(self, key: str) -> Any:
        """Enable dot notation access for dictionary keys."""
        if key in self:
            return self[key]
        raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{key}'")

    def __setattr__(self, key: str, value: Any) -> None:
        """Enable dot notation assignment for dictionary keys."""
        # Convert dictionaries to DotDict when setting attributes
        if isinstance(value, dict) and not isinstance(value, DotDict):
            value = DotDict(value)
        elif isinstance(value, list):
            # Process lists when setting them as attributes
            value = self._process_list_items(value, key)
        self[key] = value

    def __delattr__(self, key: str) -> None:
        """Enable dot notation deletion for dictionary keys."""
        if key in self:
            del self[key]
        else:
            raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{key}'")


# --------------------------------------------------------------------------------------------------


def compare_dictionaries(
    reference: dict[str, Any], candidate: dict[str, Any]
) -> tuple[list[str], list[str], bool]:
    """
    Compares two dictionaries based on their keys, including nested keys.

    Args:
        reference: The reference dictionary.
        candidate: The dictionary to compare against the reference.

    Returns:
        A tuple containing:
            - A list of keys missing in the candidate.
            - A list of keys present in the candidate but not in the reference.
            - A boolean indicating whether the comparison failed (missing keys > 0).

    Generated with Gemini Flash 2.0, edited.
    """

    missing_keys: list[str] = []
    extra_keys: list[str] = []

    def _traverse_and_compare(
        ref_dict: dict[str, Any], cand_dict: dict[str, Any], prefix: str = ""
    ) -> None:
        for key in ref_dict:
            full_key = prefix + str(key) if prefix else str(key)  # Handle nested keys
            if key not in cand_dict:
                missing_keys.append(full_key)
            elif isinstance(ref_dict[key], dict) and isinstance(cand_dict[key], dict):
                _traverse_and_compare(
                    ref_dict[key], cand_dict[key], full_key + "."
                )  # Recurse for nested dictionaries

        for key in cand_dict:
            full_key = prefix + str(key) if prefix else str(key)
            if key not in ref_dict:
                is_nested_extra = False
                for existing_key in reference:
                    if (
                        isinstance(reference[existing_key], dict)
                        and isinstance(cand_dict[key], dict)
                        and str(key).startswith(str(existing_key) + ".")
                    ):
                        is_nested_extra = True
                if not is_nested_extra:
                    extra_keys.append(full_key)

    _traverse_and_compare(reference, candidate)

    return missing_keys, extra_keys, len(missing_keys) > 0



# SPDX-FileCopyrightText: 2025 Florian Obersteiner / KIT
# SPDX-FileContributor: Florian Obersteiner <f.obersteiner@kit.edu>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""NA helpers."""

from datetime import datetime

import numpy as np
import pandas as pd
import polars as pl

###############################################################################

MICROSECONDS_PER_SECOND = 1_000_000

###############################################################################


def naDict_2_npndarr(
    naDict,
    sel_vnames=None,
    clean_vnames=False,
    vname_delimiter=";",
    split_idx=0,
    xdtype=float,
    vdtype=float,
    vmiss=np.nan,
):
    """
    Convert variables from na1001 class instance to dictionary mapping names to numpy arrays.

    See class method for detailed doc-string.
    """
    npDict = {naDict["XNAME"].split(vname_delimiter)[0]: np.array(naDict["_X"], dtype=xdtype)}

    if not sel_vnames:
        sel_vnames = [n.split(vname_delimiter)[0] for n in naDict["_VNAME"]]

    if clean_vnames:
        sel_vnames = [k.replace(" ", "") for k in sel_vnames]

    # for each parameter, find its index in naDict['V']
    for parm in sel_vnames:
        if vname_delimiter:
            ix = [name.split(vname_delimiter)[split_idx] for name in naDict["_VNAME"]].index(parm)
        else:
            ix = naDict["_VNAME"].index(parm)
        npDict[parm] = np.array(naDict["_V"][ix], dtype=vdtype)

        # check vmiss: make sure that vmiss=0 also works by checking for type
        # boolean. Might be a bit confusing since vmiss=True would also result
        # in keeping the original values.
        if not isinstance(vmiss, bool):
            npDict[parm][np.isclose(npDict[parm], float(naDict["VMISS"][ix]))] = vmiss
        # account for VSCAL:
        npDict[parm] *= float(naDict["VSCAL"][ix])

    return npDict


###############################################################################


def naDict_2_pddf(
    naDict,
    sep_colhdr="\t",
    idx_colhdr=-1,
    clean_colnames=False,
    dtype=float,
    add_datetime_index=False,
):
    """
    Convert variables from na1001 class instance to pandas dataframe.

    See class method for detailed doc-string.
    """
    # column names for the DataFrame. NA reader does not give us column names,
    # so we need to handle them here.
    keys = naDict["_NCOM"][idx_colhdr].split(sep_colhdr)
    if clean_colnames:
        keys = [k.replace(" ", "") for k in keys]

    # begin extraction with independent variable:
    values = [np.array(naDict["_X"], dtype=dtype)]

    # include scaling factors and missing values:
    vmiss = [float(s) for s in naDict["VMISS"]]
    vscal = [float(s) for s in naDict["VSCAL"]]

    # for each variable...
    for i, v_n in enumerate(naDict["_V"]):
        # cast list of string to np.array:
        arr = np.array(v_n, dtype=dtype)
        # replace missing values with np.nan:
        arr[np.isclose(arr, vmiss[i])] = np.nan
        # add array to list of arrays:
        values.append(arr * vscal[i])

    df = pd.DataFrame.from_dict(dict(zip(keys, values, strict=False)))

    if add_datetime_index:
        df.set_index(
            pd.DatetimeIndex(
                pd.Timestamp(*naDict["DATE"]) + pd.to_timedelta(values[0], unit="s"),
                tz="UTC",
            ),
            inplace=True,
        )

    return df


###############################################################################


def naDict_2_poldf(
    naDict,
    sep_colhdr="\t",
    idx_colhdr=-1,
    clean_colnames=False,
    add_datetime=False,
    nan_to_none=False,
    _dtype=float,
):
    """
    Convert variables from na1001 class instance to polars dataframe.

    See class method for detailed doc-string.
    """
    # column names for the DataFrame. NA reader does not give us column names,
    # so we need to handle them here.
    keys = naDict["_NCOM"][idx_colhdr].split(sep_colhdr)
    if clean_colnames:
        keys = [k.replace(" ", "") for k in keys]

    # # begin extraction with independent variable:
    values = [np.array(naDict["_X"], dtype=_dtype)]

    # # include scaling factors and missing values:
    vmiss = [float(s) for s in naDict["VMISS"]]
    vscal = [float(s) for s in naDict["VSCAL"]]

    # for each variable...
    for i, v_n in enumerate(naDict["_V"]):
        # cast list of string to np.array:
        arr = np.array(v_n, dtype=_dtype)
        # replace missing values with np.nan:
        arr[np.isclose(arr, vmiss[i])] = np.nan
        # add array to list of arrays:
        values.append(arr * vscal[i])

    df = pl.DataFrame(dict(zip(keys, values, strict=False)))

    if add_datetime:
        df = df.with_columns(
            (
                datetime(*naDict["DATE"])  # this will give µs resolution by default
                + pl.duration(microseconds=pl.col(keys[0]) * MICROSECONDS_PER_SECOND)
            ).alias("datetime")
        )

    if nan_to_none:
        return df.fill_nan(None)

    return df



# SPDX-FileCopyrightText: 2025 Florian Obersteiner / KIT
# SPDX-FileContributor: Florian Obersteiner <f.obersteiner@kit.edu>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import unittest
from io import BytesIO, StringIO
from pathlib import Path

import polars as pl

from pyfuppes.na1001 import FFI1001 as na1001

try:
    wd = Path(__file__).parent
except NameError:
    wd = Path.cwd()
assert wd.is_dir(), "faild to obtain working directory"

src, dst = wd / "test_input", wd / "test_output"


class TestNa1001(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # to run before all tests
        pass

    @classmethod
    def tearDownClass(cls):
        # to run after all tests
        pass

    def setUp(self):
        # to run before each test
        pass

    def tearDown(self):
        # to run after each test
        pass

    def test_read_valid(self):
        file = src / "validate_na/valid_1001a.na"
        na = na1001(file, sep_data=" ", rmv_repeated_seps=True)
        self.assertIsNotNone(na)
        s = na.__repr__()
        self.assertTrue(len(s) > 10)
        self.assertEqual(na.NV, 2)
        self.assertEqual(na.NLHEAD, 36)
        self.assertEqual(len(na._HEADER), na.NLHEAD)

        file = src / "validate_na/OM_20200304_591_CPT_MUC_V01_valid0.txt"
        na = na1001(file, sep_data="\t")
        self.assertEqual(len(na._HEADER), na.NLHEAD)
        file = src / "validate_na/OM_20200304_591_CPT_MUC_V01_valid1.txt"
        na = na1001(file, sep_data="\t")
        self.assertEqual(len(na._HEADER), na.NLHEAD)

    def test_invalid_read_config(self):
        file = src / "validate_na/valid_1001a.na"
        with self.assertRaises(
            (
                AssertionError,
                ValueError,
            )
        ):
            na1001(file, rmv_repeated_seps=False)

    def test_validate(self):
        # cases
        # (filepath, expect_fail?)
        cases = sorted((f, "invalid" in f.name) for f in (src / "validate_na").glob("*.txt"))
        self.assertNotEqual([], cases)
        for path, expect_fail in cases:
            if expect_fail:
                with self.assertRaises((AssertionError, ValueError)):
                    _ = na1001(path)
            else:
                result = na1001(path)
                self.assertIsNotNone(result)

    def test_read_from_io(self):
        s = """37 1001
F. Obersteiner; A. Zahn; f.obersteiner@kit.edu; andreas.zahn@kit.edu
Institute for Meteorology and Climate Research (IMK), Karlsruhe Institute of Technology (KIT), 76021 Karlsruhe, P.O. Box 3640, Germany
Ozone measured with the dual-beam UV-photometer that is part of the FAIRO instrument during IAGOS-CARIBIC (onboard A340-600 D-AIHE of Lufthansa)
IAGOS-CARIBIC (CARIBIC-2), http://www.caribic-atmospheric.com/
1 1
2020 03 04 2020 09 22
0
TimeCRef; CARIBIC_reference_time_since_0_hours_UTC_on_first_date_in_line_7; [s]
1
1
9999
Ozone; Ozone volume mixing ratio; [ppb]
8
FileFormatVersionInfo: CARIBIC_NAmes_v02 standard # This file was created according to the conventions for CARIBIC NASA Ames data files, given by 'Info_Nasa_Ames_Format_02.zip'.
FlightNo: 591
FlightRoute: CPT MUC # IATA code of DepartureAirport and ArrivalAirport
DepartureAirport: Cape Town
ArrivalAirport: Munich
FileName: OM_20200304_591_CPT_MUC_V01.txt
ExceptionToConsider: no
ChangeLog: 0
14
FAIRO eval software version: 2019.11.21a
Info on adjustments of the independent variable:
Measurement time-stamp is synchronized to CARIBIC master PC time.
Info on measurement uncertainty:
Uncertainty is estimated to be at least 1 ppb OR 2.0% for the UV photometer and 2.5% combined for the chemiluminescence detector (use whatever value is larger).
#
The following two lines contain standard names and units according to the 'NetCDF Climate and Forecast (CF) Metadata Convention', see http://cfconventions.org/. Each entry (name or unit) corresponds to a certain column; the entry before the first tab belongs to the independent column.
CF_StandardNames: time	mole_fraction_of_ozone_in_air
CF_Units: [s]	[1e-9]
The following four lines contain: Start date like in line 7 (formatted as YYYYMMDD), followed by the multiplicators like in line 11; Line number containing the KeyLabel 'ExceptionToConsider' (0 if no such exception is mentioned), followed by the NaN-values like in line 12; Units like in lines 13ff; Column headers.
20200304	1
0	9999
[s]	[ppb]
TimeCRef	Ozone"""
        with self.assertRaises(AssertionError):
            _ = na1001(StringIO(s))  # header-only not allowed by default

        na = na1001(StringIO(s), allow_emtpy_data=True)
        self.assertEqual(len(na._HEADER), na.NLHEAD)

        na = na1001(BytesIO(bytes(s, "utf-8")), allow_emtpy_data=True)
        self.assertEqual(len(na._HEADER), na.NLHEAD)

    def test_poldf(self):
        file = src / "validate_na/OM_20200304_591_CPT_MUC_V01_valid0.txt"
        na = na1001(file, sep_data="\t")
        df = na.to_poldf(add_datetime=True)
        self.assertTrue(isinstance(df, pl.DataFrame))
        self.assertTrue("datetime" in df.columns)


if __name__ == "__main__":
    unittest.main()



# SPDX-FileCopyrightText: 2025 Florian Obersteiner / KIT
# SPDX-FileContributor: Florian Obersteiner <f.obersteiner@kit.edu>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Text file to dictionary."""

from pathlib import Path
from typing import NamedTuple

###############################################################################


def txt_2_dict_basic(
    file: str | Path, delimiter: str, *, offset: int = 0, encoding: str = "utf-8"
) -> dict[str, list[str]]:
    """
    Most basic csv reader (delimiter-separated text file).

    Faster than dict reader from csv package.

    Input:
    file - path and filename (string or pathlib.Path)
    delimiter - line separator (string)
    offset - lines to skip at beginning (integer)
    encoding - encoding of the text file to read. default is UTF-8.
               set to None to use the operating system default.

    Returns:
        dict; keys = values from the first row, values = rest of the csv file.
    """
    with open(file, encoding=encoding) as csvfile:
        data = csvfile.read().splitlines()
    if offset > 0:
        data = data[offset:]

    separated = [[v for v in line.split(delimiter) if v] for line in data]

    return {item[0]: list(item[1:]) for item in zip(*separated, strict=False)}


###############################################################################


class CsvData(NamedTuple):
    src: str
    file_hdr: list[str]
    data: dict[str, list[str]]


def txt_2_dict(  # noqa: C901
    file: str | Path,
    delimiter: str = ";",
    colhdr_ix: int = 0,
    encoding: str = "utf-8",
    ignore_repeated_sep: bool = False,
    ignore_colhdr: bool = False,
    keys_upper: bool = False,
    preserve_empty: bool = True,
    skip_empty_lines: bool = False,
) -> CsvData:
    """
    Read csv with header.

    Input: txt file with column header and values separated by a specific separator (delimiter).

    file                - full path to txt file.
    delimiter           - value separator (delimiter), e.g. ";" in a csv file.
    colhdr_ix           - row index of the column header.
    encoding            - encoding of the text file to read. default is UTF-8.
                          set to None to use the operating system default.
    ignore_repeated_sep - if set to True, repeated occurrences of delimiter are
                          ignored during extraction of elements from the
                          file lines
                          Warning: empty fields must then be filled with a
                          "no-value indicator" (e.g. string NULL)
    keys_upper          - convert key name (from column header) to upper-case
    preserve_empty      - do not remove empty fields
    skip_empty_lines    - ignore empty lines, just skip them

    Returns:
      NamedTuple 'CsvData';
        .src      - file path, str
        .file_hdr - file header (if any), list of str
        .data     - dict[str list[str]]
    """

    with open(file, encoding=encoding) as file_obj:
        content = file_obj.readlines()

    if not content:
        raise ValueError(f"no content in {file}")  # noqa: TRY003

    result = {"file_hdr": [], "data": {}, "src": file if isinstance(file, str) else file.as_posix()}

    if colhdr_ix > 0:
        result["file_hdr"] = [line.strip() for line in content[:colhdr_ix]]

    col_hdr = content[colhdr_ix].strip().rsplit(delimiter)
    if ignore_repeated_sep:
        col_hdr = [s for s in col_hdr if s != ""]
    if ignore_colhdr:
        for i, _ in enumerate(col_hdr):
            col_hdr[i] = f"col_{(i + 1):03d}"

    if keys_upper:
        col_hdr = [s.upper() for s in col_hdr]

    for element in col_hdr:
        result["data"][element] = []  # type: ignore

    if ignore_colhdr:  # cut col header...
        colhdr_ix -= 1

    content = content[1 + colhdr_ix :]
    for ix, ln in enumerate(content):
        line = ln.strip("\n")
        if not preserve_empty:
            line = line.strip(delimiter)

        if skip_empty_lines and line == "":  # skip empty lines
            continue

        _line = line.rsplit(delimiter)

        if ignore_repeated_sep:
            _line = [s for s in line if s != ""]

        if len(_line) != len(col_hdr):
            err_msg = f"{len(_line)} elements in line {ix + 1 + colhdr_ix} != {len(col_hdr)} elem in col header ({file}).\nLine content: {_line}"
            raise ValueError(err_msg)

        for i, hdr_tag in enumerate(col_hdr):
            result["data"][hdr_tag].append(_line[i].strip())  # type: ignore

    return CsvData(result["src"], result["file_hdr"], result["data"])  # type: ignore  # noqa: PGH003
